"""
Copyright 2021 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
from unittest import mock

import pytest
from pandas._testing import assert_frame_equal

from gs_quant.common import Currency
from gs_quant.models.risk_model import FactorRiskModel, MacroRiskModel, ReturnFormat, Unit
from gs_quant.models.risk_model_utils import get_optional_data_as_dataframe, _map_measure_to_field_name
from gs_quant.session import GsSession, Environment
from gs_quant.target.risk_models import RiskModel as Risk_Model, RiskModelCoverage, RiskModelTerm, \
    RiskModelUniverseIdentifier, RiskModelType, RiskModelDataAssetsRequest as DataAssetsRequest, \
    RiskModelDataMeasure as Measure, RiskModelUniverseIdentifierRequest as UniverseIdentifier

empty_entitlements = {
    "execute": [],
    "edit": [],
    "view": [],
    "admin": [],
    "query": [],
    "upload": []
}

mock_risk_model_obj = Risk_Model(RiskModelCoverage.Country,
                                 'model_id',
                                 'Fake Risk Model',
                                 RiskModelTerm.Long,
                                 RiskModelUniverseIdentifier.gsid,
                                 'GS',
                                 1.0,
                                 universe_size=10000,
                                 entitlements=empty_entitlements,
                                 description='Test',
                                 expected_update_time='00:00:00',
                                 type=RiskModelType.Factor
                                 )

mock_macro_risk_model_obj = Risk_Model(coverage=RiskModelCoverage.Country,
                                       id='macro_model_id',
                                       name='Fake Risk Model',
                                       term=RiskModelTerm.Long,
                                       universe_identifier=RiskModelUniverseIdentifier.gsid,
                                       vendor='GS',
                                       version=1.0,
                                       entitlements=empty_entitlements,
                                       description='Test',
                                       expected_update_time='00:00:00',
                                       type=RiskModelType.Macro
                                       )


def mock_risk_model(mocker):
    from gs_quant.session import OAuth2Session
    OAuth2Session.init = mock.MagicMock(return_value=None)
    GsSession.use(Environment.QA, 'client_id', 'secret')
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_post', return_value=mock_risk_model_obj)
    mocker.patch.object(GsSession.current, '_get', return_value=mock_risk_model_obj)
    mocker.patch.object(GsSession.current, '_put', return_value=mock_risk_model_obj)
    return FactorRiskModel.get('model_id')


def mock_macro_risk_model(mocker):
    from gs_quant.session import OAuth2Session
    OAuth2Session.init = mock.MagicMock(return_value=None)
    GsSession.use(Environment.QA, 'client_id', 'secret')
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_post', return_value=mock_macro_risk_model_obj)
    mocker.patch.object(GsSession.current, '_get', return_value=mock_macro_risk_model_obj)
    mocker.patch.object(GsSession.current, '_put', return_value=mock_macro_risk_model_obj)
    return MacroRiskModel.get('macro_model_id')


def test_create_risk_model(mocker):
    mock_risk_model(mocker)
    risk_model_id = 'model_id'
    mocker.patch.object(GsSession.current, '_post', return_value=mock_risk_model_obj)
    new_model = FactorRiskModel(risk_model_id,
                                'Fake Risk Model',
                                RiskModelCoverage.Country,
                                RiskModelTerm.Long,
                                RiskModelUniverseIdentifier.gsid,
                                'GS',
                                0.1,
                                universe_size=10000,
                                entitlements={},
                                description='Test',
                                expected_update_time=dt.datetime.strptime('00:00:00', '%H:%M:%S').time())
    new_model.save()
    assert new_model.id == mock_risk_model_obj.id
    assert new_model.name == mock_risk_model_obj.name
    assert new_model.description == mock_risk_model_obj.description
    assert new_model.term == mock_risk_model_obj.term
    assert new_model.universe_size == mock_risk_model_obj.universe_size
    assert new_model.coverage == mock_risk_model_obj.coverage
    assert new_model.universe_identifier == mock_risk_model_obj.universe_identifier
    assert new_model.expected_update_time == dt.datetime.strptime(
        mock_risk_model_obj.expected_update_time, '%H:%M:%S').time()


def test_update_risk_model_entitlements(mocker):
    new_model = mock_risk_model(mocker)
    new_entitlements = {
        "execute": ['guid:X'],
        "edit": [],
        "view": [],
        "admin": [],
        "query": [],
        "upload": []
    }

    new_model.entitlements = new_entitlements
    new_model.save()
    assert 'guid:X' in new_model.entitlements.get('execute')
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    new_model.entitlements = empty_entitlements
    new_model.save()
    new_entitlements = {
        "execute": ['guid:X'],
        "edit": [],
        "view": [],
        "admin": ['guid:XX'],
        "query": [],
        "upload": ['guid:XXX']
    }
    new_model.entitlements = new_entitlements
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert 'guid:X' in new_model.entitlements.get('execute')
    assert 'guid:XX' in new_model.entitlements.get('admin')
    assert 'guid:XXX' in new_model.entitlements.get('upload')


def test_update_risk_model(mocker):
    new_model = mock_risk_model(mocker)

    new_model.term = RiskModelTerm.Short
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.term == RiskModelTerm.Short

    new_model.description = 'Test risk model'
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.description == 'Test risk model'

    new_model.vendor = 'GS'
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.vendor == 'GS'

    new_model.term = RiskModelTerm.Medium
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.term == RiskModelTerm.Medium

    new_model.version = 0.1
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.version == 0.1

    new_model.universe_size = 10000
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.universe_size == 10000

    new_model.coverage = RiskModelCoverage.Global
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.coverage == RiskModelCoverage.Global

    new_model.name = 'TEST RISK MODEL'
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.name == 'TEST RISK MODEL'

    new_model.expected_update_time = dt.time(1, 0, 0)
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.expected_update_time == dt.time(1, 0, 0)

    new_model.type = RiskModelType.Thematic
    new_model.save()
    mocker.patch.object(GsSession.current, '_get', return_value=new_model)
    assert new_model.type == RiskModelType.Thematic


def test_get_r_squared(mocker):
    macro_model = mock_macro_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.R_Squared, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["904026", "232128", "24985", "160444"],
                    "rSquared": [89.0, 45.0, 12.0, 5.0]
                }
            }
        ],
        'totalResults': 1
    }

    r_squared_response = {
        '160444': {'2022-04-05': 5.0},
        '232128': {'2022-04-05': 45.0},
        '24985': {'2022-04-05': 12.0},
        '904026': {'2022-04-05': 89.0}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = macro_model.get_r_squared(start_date=dt.date(2022, 4, 4),
                                         end_date=dt.date(2022, 4, 6),
                                         assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                         format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='macro_model_id'),
                                               query, timeout=200)
    assert response == r_squared_response


def test_get_fair_value_gap_standard_deviation(mocker):
    macro_model = mock_macro_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.Fair_Value_Gap_Standard_Deviation, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["904026", "232128", "24985", "160444"],
                    "fairValueGapStandardDeviation": [4.0, 5.0, 1.0, 7.0]
                }
            }
        ],
        'totalResults': 1
    }

    fvg_response = {
        '160444': {'2022-04-05': 7.0},
        '232128': {'2022-04-05': 5.0},
        '24985': {'2022-04-05': 1.0},
        '904026': {'2022-04-05': 4.0}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)
    response = macro_model.get_fair_value_gap(start_date=dt.date(2022, 4, 4),
                                              end_date=dt.date(2022, 4, 6),
                                              assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                              format=ReturnFormat.JSON)
    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='macro_model_id'),
                                               query, timeout=200)
    assert response == fvg_response


def test_get_fair_value_gap_percent(mocker):
    macro_model = mock_macro_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.Fair_Value_Gap_Percent, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["904026", "232128", "24985", "160444"],
                    "fairValueGapPercent": [90.0, 34.0, 8.0, 34.0]
                }
            }
        ],
        'totalResults': 1
    }

    fvg_response = {
        '160444': {'2022-04-05': 34.0},
        '232128': {'2022-04-05': 34.0},
        '24985': {'2022-04-05': 8.0},
        '904026': {'2022-04-05': 90.0}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)
    response = macro_model.get_fair_value_gap(start_date=dt.date(2022, 4, 4),
                                              end_date=dt.date(2022, 4, 6),
                                              assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                              fair_value_gap_unit=Unit.PERCENT,
                                              format=ReturnFormat.JSON)
    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='macro_model_id'),
                                               query, timeout=200)
    assert response == fvg_response


@pytest.mark.parametrize("statistical_measure, fieldKey", [
    (Measure.Factor_Mean, _map_measure_to_field_name(Measure.Factor_Mean)),
    (Measure.Factor_Cross_Sectional_Mean, _map_measure_to_field_name(Measure.Factor_Cross_Sectional_Mean)),
    (Measure.Factor_Standard_Deviation, _map_measure_to_field_name(Measure.Factor_Standard_Deviation)),
    (Measure.Factor_Cross_Sectional_Standard_Deviation,
     _map_measure_to_field_name(Measure.Factor_Cross_Sectional_Standard_Deviation))
])
def test_get_statistical_factor_data(mocker, statistical_measure, fieldKey):
    risk_model = mock_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [statistical_measure, Measure.Factor_Name, Measure.Factor_Id,
                     Measure.Universe_Factor_Exposure, Measure.Asset_Universe],
        'limitFactors': True
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "factorData": [
                    {
                        "factorId": "1",
                        f"{fieldKey}": 89.0,
                        "factorName": "Factor1"
                    },
                    {
                        "factorId": "2",
                        f"{fieldKey}": 0.67,
                        "factorName": "Factor2"
                    }
                ],
                'assetData': {
                    'factorExposure': [{'1': 0.2, '2': 0.3}, {'1': 0.02, '2': 0.03},
                                       {'1': 6.2, '2': 3.0}, {'1': -6.2, '2': 0.3}],
                    'universe': ['904026', '232128', '24985', '160444']
                }
            }
        ],
        'totalResults': 1
    }

    expected_response = {
        'Factor1': {'2022-04-05': 89.0},
        'Factor2': {'2022-04-05': 0.67}
    }

    kwargs = {
        "start_date": dt.date(2022, 4, 4),
        "end_date": dt.date(2022, 4, 6),
        "assets": DataAssetsRequest(UniverseIdentifier.gsid, universe),
        "format": ReturnFormat.JSON
    }

    field_key_to_getter_ref = {
        _map_measure_to_field_name(Measure.Factor_Mean): risk_model.get_factor_mean,
        _map_measure_to_field_name(Measure.Factor_Cross_Sectional_Mean): risk_model.get_factor_cross_sectional_mean,
        _map_measure_to_field_name(Measure.Factor_Standard_Deviation): risk_model.get_factor_standard_deviation,
        _map_measure_to_field_name(
            Measure.Factor_Cross_Sectional_Standard_Deviation): risk_model.get_factor_cross_sectional_standard_deviation
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    actual_response = field_key_to_getter_ref[fieldKey](**kwargs)
    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)

    assert actual_response == expected_response


def test_get_factor_z_score(mocker):
    macro_model = mock_macro_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.Factor_Z_Score, Measure.Factor_Name, Measure.Factor_Id,
                     Measure.Universe_Factor_Exposure, Measure.Asset_Universe],
        'limitFactors': True
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "factorData": [
                    {
                        "factorId": "1",
                        "factorZScore": 1.5,
                        "factorName": "Factor1"
                    },
                    {
                        "factorId": "2",
                        "factorZScore": -1.0,
                        "factorName": "Factor2"
                    }
                ],
                'assetData': {
                    'factorExposure': [{'1': 0.2, '2': 0.3}, {'1': 0.02, '2': 0.03},
                                       {'1': 6.2, '2': 3.0}, {'1': -6.2, '2': 0.3}],
                    'universe': ['904026', '232128', '24985', '160444']
                }
            }
        ],
        'totalResults': 1
    }

    factor_z_score_response = {
        'Factor1': {'2022-04-05': 1.5},
        'Factor2': {'2022-04-05': -1.0}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = macro_model.get_factor_z_score(start_date=dt.date(2022, 4, 4),
                                              end_date=dt.date(2022, 4, 6),
                                              assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                              format=ReturnFormat.JSON)
    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='macro_model_id'),
                                               query, timeout=200)
    assert response == factor_z_score_response


def test_get_predicted_beta(mocker):
    model = mock_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.Predicted_Beta, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["904026", "232128", "24985", "160444"],
                    "predictedBeta": [0.4, 1.5, 1.2, 0.5]
                }
            }
        ],
        'totalResults': 1
    }

    predicted_beta_response = {
        '160444': {'2022-04-05': 0.5},
        '232128': {'2022-04-05': 1.5},
        '24985': {'2022-04-05': 1.2},
        '904026': {'2022-04-05': 0.4}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_predicted_beta(start_date=dt.date(2022, 4, 4),
                                        end_date=dt.date(2022, 4, 6),
                                        assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                        format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == predicted_beta_response


def test_get_global_predicted_beta(mocker):
    model = mock_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.Global_Predicted_Beta, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["904026", "232128", "24985", "160444"],
                    "globalPredictedBeta": [0.4, 1.5, 1.2, 0.5]
                }
            }
        ],
        'totalResults': 1
    }

    global_predicted_beta_response = {
        '160444': {'2022-04-05': 0.5},
        '232128': {'2022-04-05': 1.5},
        '24985': {'2022-04-05': 1.2},
        '904026': {'2022-04-05': 0.4}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_global_predicted_beta(start_date=dt.date(2022, 4, 4),
                                               end_date=dt.date(2022, 4, 6),
                                               assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                               format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == global_predicted_beta_response


def test_get_estimation_universe_weights(mocker):
    model = mock_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.Estimation_Universe_Weight, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["904026", "232128", "24985", "160444"],
                    "estimationUniverseWeight": [0.4, None, None, 0.6]
                }
            }
        ],
        'totalResults': 1
    }

    estu_response = {
        '160444': {'2022-04-05': 0.6},
        '232128': {'2022-04-05': None},
        '24985': {'2022-04-05': None},
        '904026': {'2022-04-05': 0.4}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_estimation_universe_weights(start_date=dt.date(2022, 4, 4),
                                                     end_date=dt.date(2022, 4, 6),
                                                     assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                                     format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == estu_response


def test_get_daily_return(mocker):
    model = mock_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.Daily_Return, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["904026", "232128", "24985", "160444"],
                    "dailyReturn": [-0.4, -1.5, 1.2, 0.5]
                }
            }
        ],
        'totalResults': 1
    }

    daily_return_response = {
        '160444': {'2022-04-05': 0.5},
        '232128': {'2022-04-05': -1.5},
        '24985': {'2022-04-05': 1.2},
        '904026': {'2022-04-05': -0.4}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_daily_return(start_date=dt.date(2022, 4, 4),
                                      end_date=dt.date(2022, 4, 6),
                                      assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                      format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == daily_return_response


def test_get_specific_return(mocker):
    model = mock_risk_model(mocker)
    universe = ["904026", "232128", "24985", "160444"]
    query = {
        'startDate': '2022-04-04',
        'endDate': '2022-04-06',
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, universe),
        'measures': [Measure.Specific_Return, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'missingDates': ['2022-04-04', '2022-04-06'],
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["904026", "232128", "24985", "160444"],
                    "specificReturn": [0.5, 1.6, 1.4, 0.7]
                }
            }
        ],
        'totalResults': 1
    }

    specific_return_response = {
        '160444': {'2022-04-05': 0.7},
        '232128': {'2022-04-05': 1.6},
        '24985': {'2022-04-05': 1.4},
        '904026': {'2022-04-05': 0.5}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_specific_return(start_date=dt.date(2022, 4, 4),
                                         end_date=dt.date(2022, 4, 6),
                                         assets=DataAssetsRequest(UniverseIdentifier.gsid, universe),
                                         format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == specific_return_response


@pytest.mark.parametrize("aws_upload", [True, False])
def test_upload_risk_model_data(mocker, aws_upload):
    model = mock_risk_model(mocker)
    risk_model_data = {
        'date': '2023-04-14',
        'assetData': {
            'universe': ['2407966', '2046251', 'USD'],
            'specificRisk': [12.09, 45.12, 3.09],
            'factorExposure': [
                {'1': 0.23, '2': 0.023},
                {'1': 0.23},
                {'3': 0.23, '2': 0.023}
            ],
            'totalRisk': [0.12, 0.45, 1.2]
        },
        'factorData': [
            {
                'factorId': '1',
                'factorName': 'USD',
                'factorCategory': 'Currency',
                'factorCategoryId': 'CUR'
            },
            {
                'factorId': '2',
                'factorName': 'ST',
                'factorCategory': 'ST',
                'factorCategoryId': 'ST'
            },
            {
                'factorId': '3',
                'factorName': 'IND',
                'factorCategory': 'IND',
                'factorCategoryId': 'IND'
            }
        ],
        'covarianceMatrix': [[0.089, 0.0123, 0.345],
                             [0.0123, 3.45, 0.345],
                             [0.345, 0.345, 1.23]],
        'issuerSpecificCovariance': {
            'universeId1': ['2407966'],
            'universeId2': ['2046251'],
            'covariance': [0.03754]
        },
        'factorPortfolios': {
            'universe': ['2407966', '2046251'],
            'portfolio': [{'factorId': 1, 'weights': [0.25, 0.75]},
                          {'factorId': 2, 'weights': [0.25, 0.75]},
                          {'factorId': 3, 'weights': [0.25, 0.75]}]
        }
    }

    base_url = f"/risk/models/data/{model.id}?partialUpload=true"
    date = risk_model_data.get("date")
    max_asset_batch_size = 2

    batched_asset_data = [
        {"assetData": {key: value[i:i + max_asset_batch_size] for key, value in
                       risk_model_data.get("assetData").items()}, "date": date,
         } for i in range(0, len(risk_model_data.get("assetData").get("universe")), max_asset_batch_size)
    ]

    max_asset_batch_size //= 2
    batched_factor_portfolios = [
        {"factorPortfolios": {key: (value[i:i + max_asset_batch_size] if key in "universe" else
                                    [{"factorId": factor_weights.get("factorId"),
                                      "weights": factor_weights.get("weights")[i:i + max_asset_batch_size]} for
                                     factor_weights in
                                     value])
                              for key, value in risk_model_data.get("factorPortfolios").items()},
         "date": date
         } for i in range(0, len(risk_model_data.get("factorPortfolios").get("universe")), max_asset_batch_size)
    ]

    expected_factor_data_calls = [
        mock.call(f"{base_url}{'&awsUpload=true' if aws_upload else ''}",
                  {"date": date, "factorData": risk_model_data.get("factorData"),
                   "covarianceMatrix": risk_model_data.get("covarianceMatrix")}, timeout=200)
    ]

    expected_asset_data_calls = []
    for batch_num, batch_asset_payload in enumerate(batched_asset_data):
        final_upload_flag = 'true' if batch_num == len(batched_asset_data) - 1 else 'false'
        expected_asset_data_calls.append(
            mock.call(f"{base_url}&finalUpload={final_upload_flag}{'&awsUpload=true' if aws_upload else ''}",
                      batch_asset_payload, timeout=200)
        )

    expected_factor_portfolios_data_calls = []
    for batch_num, batched_fp_payload in enumerate(batched_factor_portfolios):
        final_upload_flag = 'true' if batch_num == len(batched_factor_portfolios) - 1 else 'false'
        expected_factor_portfolios_data_calls.append(
            mock.call(f"{base_url}&finalUpload={final_upload_flag}{'&awsUpload=true' if aws_upload else ''}",
                      batched_fp_payload, timeout=200)
        )

    expected_isc_data_calls = [
        mock.call(f"{base_url}&finalUpload=true{'&awsUpload=true' if aws_upload else ''}",
                  {"issuerSpecificCovariance": risk_model_data.get("issuerSpecificCovariance"), "date": date},
                  timeout=200)
    ]

    expected_calls = (expected_factor_data_calls + expected_asset_data_calls + expected_isc_data_calls +
                      expected_factor_portfolios_data_calls)

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_post', return_value='Upload Successful')

    max_asset_batch_size = 2
    model.upload_data(risk_model_data, max_asset_batch_size=max_asset_batch_size, aws_upload=aws_upload)

    call_args_list = GsSession.current._post.call_args_list

    assert len(call_args_list) == len(expected_calls)
    assert call_args_list == expected_calls

    GsSession.current._post.assert_has_calls(expected_calls, any_order=False)


@pytest.mark.parametrize("days", [0, 30, 60, 90])
def test_get_bid_ask_spread(mocker, days):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    measure_to_query = Measure.Bid_Ask_Spread if not days else Measure[f'Bid_Ask_Spread_{days}d']
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [measure_to_query, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    f"bidAskSpread{'' if days == 0 else f'{days}d'}": [0.5, 1.6]
                }
            }
        ],
        'totalResults': 1
    }

    bid_ask_spread_response = {
        '2588173': {'2022-04-05': 1.6},
        '2046251': {'2022-04-05': 0.5}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_bid_ask_spread(start_date=dt.date(2022, 4, 5),
                                        end_date=dt.date(2022, 4, 5),
                                        days=days,
                                        assets=assets,
                                        format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == bid_ask_spread_response


@pytest.mark.parametrize("days", [0, 30, 60, 90])
def test_get_trading_volume(mocker, days):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    measure_to_query = Measure.Trading_Volume if not days else Measure[f'Trading_Volume_{days}d']
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [measure_to_query, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    f"tradingVolume{'' if days == 0 else f'{days}d'}": [0.5, 1.6]
                }
            }
        ],
        'totalResults': 1
    }

    trading_volume_response = {
        '2588173': {'2022-04-05': 1.6},
        '2046251': {'2022-04-05': 0.5}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_trading_volume(start_date=dt.date(2022, 4, 5),
                                        end_date=dt.date(2022, 4, 5),
                                        days=days,
                                        assets=assets,
                                        format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == trading_volume_response


@pytest.mark.parametrize("days", [0, 30])
def test_get_traded_value(mocker, days):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    measure_to_query = Measure.Traded_Value_30d if not days else Measure[f'Traded_Value_{days}d']
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [measure_to_query, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "tradedValue30d": [0.5, 1.6]
                }
            }
        ],
        'totalResults': 1
    }

    trading_value_30d_response = {
        '2588173': {'2022-04-05': 1.6},
        '2046251': {'2022-04-05': 0.5}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_traded_value(start_date=dt.date(2022, 4, 5),
                                      end_date=dt.date(2022, 4, 5),
                                      days=days,
                                      assets=assets,
                                      format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == trading_value_30d_response


@pytest.mark.parametrize("days", [0, 30, 60, 90])
def test_get_composite_volume(mocker, days):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    measure_to_query = Measure.Composite_Volume if not days else Measure[f'Composite_Volume_{days}d']
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [measure_to_query, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    f"compositeVolume{'' if days == 0 else f'{days}d'}": [0.5, 1.6]
                }
            }
        ],
        'totalResults': 1
    }

    composite_volume_response = {
        '2588173': {'2022-04-05': 1.6},
        '2046251': {'2022-04-05': 0.5}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_composite_volume(start_date=dt.date(2022, 4, 5),
                                          end_date=dt.date(2022, 4, 5),
                                          days=days,
                                          assets=assets,
                                          format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == composite_volume_response


@pytest.mark.parametrize("days", [0, 30])
def test_get_composite_value(mocker, days):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    measure_to_query = Measure.Composite_Value_30d if not days else Measure[f'Composite_Value_{days}d']
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [measure_to_query, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "compositeValue30d": [0.5, 1.6]
                }
            }
        ],
        'totalResults': 1
    }

    composite_value_response = {
        '2588173': {'2022-04-05': 1.6},
        '2046251': {'2022-04-05': 0.5}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_composite_value(start_date=dt.date(2022, 4, 5),
                                         end_date=dt.date(2022, 4, 5),
                                         days=days,
                                         assets=assets,
                                         format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == composite_value_response


def test_get_issuer_market_cap(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Issuer_Market_Cap, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "issuerMarketCap": [1000000000, 2000000000]
                }
            }
        ],
        'totalResults': 1
    }

    issuer_market_cap_response = {
        '2588173': {'2022-04-05': 2000000000},
        '2046251': {'2022-04-05': 1000000000}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_issuer_market_cap(start_date=dt.date(2022, 4, 5),
                                           end_date=dt.date(2022, 4, 5),
                                           assets=assets,
                                           format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == issuer_market_cap_response


def test_get_asset_price(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Price, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "price": [100, 200]
                }
            }
        ],
        'totalResults': 1
    }

    price_response = {
        '2588173': {'2022-04-05': 200},
        '2046251': {'2022-04-05': 100}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_asset_price(start_date=dt.date(2022, 4, 5),
                                     end_date=dt.date(2022, 4, 5),
                                     assets=assets,
                                     format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == price_response


def test_get_asset_capitalization(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Capitalization, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "capitalization": [1000000000, 2000000000]
                }
            }
        ],
        'totalResults': 1
    }

    capitalization_response = {
        '2588173': {'2022-04-05': 2000000000},
        '2046251': {'2022-04-05': 1000000000}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_asset_capitalization(start_date=dt.date(2022, 4, 5),
                                              end_date=dt.date(2022, 4, 5),
                                              assets=assets,
                                              format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == capitalization_response


def test_get_currency(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Currency, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "currency": ["USD", "GBP"]
                }
            }
        ],
        'totalResults': 1
    }

    currency_response = {
        '2588173': {'2022-04-05': "GBP"},
        '2046251': {'2022-04-05': "USD"}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_currency(start_date=dt.date(2022, 4, 5),
                                  end_date=dt.date(2022, 4, 5),
                                  assets=assets,
                                  format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == currency_response


def test_get_unadjusted_specific_risk(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Unadjusted_Specific_Risk, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "unadjustedSpecificRisk": [0.5, 1.6]
                }
            }
        ],
        'totalResults': 1
    }

    unadjusted_specific_risk_response = {
        '2588173': {'2022-04-05': 1.6},
        '2046251': {'2022-04-05': 0.5}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_unadjusted_specific_risk(start_date=dt.date(2022, 4, 5),
                                                  end_date=dt.date(2022, 4, 5),
                                                  assets=assets,
                                                  format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == unadjusted_specific_risk_response


def test_get_dividend_yield(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Dividend_Yield, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "dividendYield": [0.5, 1.6]
                }
            }
        ],
        'totalResults': 1
    }

    dividend_yield_response = {
        '2588173': {'2022-04-05': 1.6},
        '2046251': {'2022-04-05': 0.5}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_dividend_yield(start_date=dt.date(2022, 4, 5),
                                        end_date=dt.date(2022, 4, 5),
                                        assets=assets,
                                        format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == dividend_yield_response


def test_get_model_price(mocker):
    model = mock_macro_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Model_Price, Measure.Asset_Universe],
        'limitFactors': False
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "assetData": {
                    "universe": ["2046251", "2588173"],
                    "modelPrice": [100, 200]
                }
            }
        ],
        'totalResults': 1
    }

    model_price_response = {
        '2588173': {'2022-04-05': 200},
        '2046251': {'2022-04-05': 100}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_model_price(start_date=dt.date(2022, 4, 5),
                                     end_date=dt.date(2022, 4, 5),
                                     assets=assets,
                                     format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='macro_model_id'),
                                               query, timeout=200)
    assert response == model_price_response


def test_get_covariance_matrix(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Covariance_Matrix, Measure.Factor_Name, Measure.Factor_Id,
                     Measure.Universe_Factor_Exposure,
                     Measure.Asset_Universe],
        'limitFactors': True
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "factorData": [
                    {
                        "factorId": "1",
                        "factorName": "factor1",
                    },
                    {
                        "factorId": "2",
                        "factorName": "factor2",
                    },
                    {
                        "factorId": "3",
                        "factorName": "factor3",
                    },
                    {
                        "factorId": "4",
                        "factorName": "factor4",
                    }
                ],
                "covarianceMatrix": [[0.5, 0.6, 0.7, 0.8], [0.3, 0.7, 0.8, 0.9], [0.5, 0.6, 0.7, 0.8],
                                     [0.3, 0.7, 0.8, 0.9]]
            }
        ],
        'totalResults': 1
    }

    covariance_matrix_response = [{
        "date": "2022-04-05",
        "factorData": [
            {
                "factorId": "1",
                "factorName": "factor1",
            },
            {
                "factorId": "2",
                "factorName": "factor2",
            },
            {
                "factorId": "3",
                "factorName": "factor3",
            },
            {
                "factorId": "4",
                "factorName": "factor4",
            }
        ],
        "covarianceMatrix": [[0.5, 0.6, 0.7, 0.8], [0.3, 0.7, 0.8, 0.9], [0.5, 0.6, 0.7, 0.8],
                             [0.3, 0.7, 0.8, 0.9]]
    }]

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_covariance_matrix(start_date=dt.date(2022, 4, 5),
                                           end_date=dt.date(2022, 4, 5),
                                           assets=assets,
                                           format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == covariance_matrix_response


def test_get_unadjusted_covariance_matrix(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Unadjusted_Covariance_Matrix, Measure.Factor_Name, Measure.Factor_Id,
                     Measure.Universe_Factor_Exposure, Measure.Asset_Universe],
        'limitFactors': True
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "factorData": [
                    {
                        "factorId": "1",
                        "factorName": "factor1",
                    },
                    {
                        "factorId": "2",
                        "factorName": "factor2",
                    },
                    {
                        "factorId": "3",
                        "factorName": "factor3",
                    },
                    {
                        "factorId": "4",
                        "factorName": "factor4",
                    }
                ],
                "unadjustedCovarianceMatrix": [[0.5, 0.6, 0.7, 0.8], [0.3, 0.7, 0.8, 0.9], [0.5, 0.6, 0.7, 0.8],
                                               [0.3, 0.7, 0.8, 0.9]]
            }
        ],
        'totalResults': 1
    }

    unadjusted_covariance_matrix_response = [{
        "date": "2022-04-05",
        "factorData": [
            {
                "factorId": "1",
                "factorName": "factor1",
            },
            {
                "factorId": "2",
                "factorName": "factor2",
            },
            {
                "factorId": "3",
                "factorName": "factor3",
            },
            {
                "factorId": "4",
                "factorName": "factor4",
            }
        ],
        "unadjustedCovarianceMatrix": [[0.5, 0.6, 0.7, 0.8], [0.3, 0.7, 0.8, 0.9], [0.5, 0.6, 0.7, 0.8],
                                       [0.3, 0.7, 0.8, 0.9]]
    }]

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_unadjusted_covariance_matrix(start_date=dt.date(2022, 4, 5),
                                                      end_date=dt.date(2022, 4, 5),
                                                      assets=assets,
                                                      format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == unadjusted_covariance_matrix_response


def test_get_pre_vra_covariance_matrix(mocker):
    model = mock_risk_model(mocker)

    universe = ["2046251", "2588173"]
    assets = DataAssetsRequest(UniverseIdentifier.sedol, universe)
    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'assets': assets,
        'measures': [Measure.Pre_VRA_Covariance_Matrix, Measure.Factor_Name, Measure.Factor_Id,
                     Measure.Universe_Factor_Exposure, Measure.Asset_Universe],
        'limitFactors': True
    }

    results = {
        'results': [
            {
                "date": "2022-04-05",
                "factorData": [
                    {
                        "factorId": "1",
                        "factorName": "factor1",
                    },
                    {
                        "factorId": "2",
                        "factorName": "factor2",
                    },
                    {
                        "factorId": "3",
                        "factorName": "factor3",
                    },
                    {
                        "factorId": "4",
                        "factorName": "factor4",
                    }
                ],
                "preVRACovarianceMatrix": [[0.5, 0.6, 0.7, 0.8], [0.3, 0.7, 0.8, 0.9], [0.5, 0.6, 0.7, 0.8],
                                           [0.3, 0.7, 0.8, 0.9]]
            }
        ],
        'totalResults': 1
    }

    pre_vra_covariance_matrix_response = [{
        "date": "2022-04-05",
        "factorData": [
            {
                "factorId": "1",
                "factorName": "factor1",
            },
            {
                "factorId": "2",
                "factorName": "factor2",
            },
            {
                "factorId": "3",
                "factorName": "factor3",
            },
            {
                "factorId": "4",
                "factorName": "factor4",
            }
        ],
        "preVRACovarianceMatrix": [[0.5, 0.6, 0.7, 0.8], [0.3, 0.7, 0.8, 0.9], [0.5, 0.6, 0.7, 0.8],
                                   [0.3, 0.7, 0.8, 0.9]]
    }]

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    # run test
    response = model.get_pre_vra_covariance_matrix(start_date=dt.date(2022, 4, 5),
                                                   end_date=dt.date(2022, 4, 5),
                                                   assets=assets,
                                                   format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == pre_vra_covariance_matrix_response


def test_get_risk_free_rate(mocker):
    model = mock_risk_model(mocker)

    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'measures': [Measure.Risk_Free_Rate],
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, []),
        'limitFactors': False
    }
    currencies = ["EUR", "INR"]
    risk_free_rate = [1.08, 0.012]
    results = {
        'results': [
            {
                "date": "2022-04-05",
                "currencyRatesData": {
                    "riskFreeRate": risk_free_rate,
                    "currency": currencies,
                },
            }
        ],
        'totalResults': 1
    }

    risk_free_rate_response = {
        "currency": {('2022-04-05', 0): 'EUR', ('2022-04-05', 1): 'INR'},
        "riskFreeRate": {('2022-04-05', 0): 1.08, ('2022-04-05', 1): 0.012}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    response = model.get_risk_free_rate(start_date=dt.date(2022, 4, 5),
                                        end_date=dt.date(2022, 4, 5),
                                        format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == risk_free_rate_response

    # fitler risk free rates by currency
    expected_filtered_rates = {
        "currency": {('2022-04-05', 1): 'INR'},
        "riskFreeRate": {('2022-04-05', 1): 0.012}
    }
    actual_filtered_rates = model.get_risk_free_rate(start_date=dt.date(2022, 4, 5),
                                                     end_date=dt.date(2022, 4, 5),
                                                     currencies=[Currency.INR],
                                                     format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)

    assert actual_filtered_rates == expected_filtered_rates

    # test DataFrame return format
    expected_data_frame = get_optional_data_as_dataframe([
        {
            "date": "2022-04-05",
            "currencyRatesData": {
                "riskFreeRate": risk_free_rate,
                "currency": currencies,
            },
        }
    ], "currencyRatesData")

    expected_data_frame = expected_data_frame.loc[expected_data_frame['currency'].isin(['INR'])]
    actual_data_frame = model.get_risk_free_rate(start_date=dt.date(2022, 4, 5),
                                                 end_date=dt.date(2022, 4, 5),
                                                 currencies=[Currency.INR])

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)

    assert_frame_equal(expected_data_frame, actual_data_frame, check_like=True)


def test_get_currency_exchange_rate(mocker):
    model = mock_risk_model(mocker)

    query = {
        'startDate': '2022-04-05',
        'endDate': '2022-04-05',
        'measures': [Measure.Currency_Exchange_Rate],
        'assets': DataAssetsRequest(UniverseIdentifier.gsid, []),
        'limitFactors': False
    }
    currencies = ["EUR", "INR"]
    currency_exchange_rate = [1.08, 0.012]
    results = {
        'results': [
            {
                "date": "2022-04-05",
                "currencyRatesData": {
                    "exchangeRate": currency_exchange_rate,
                    "currency": currencies,
                },
            }
        ],
        'totalResults': 1
    }

    currency_exchange_rate_response = {
        "currency": {('2022-04-05', 0): 'EUR', ('2022-04-05', 1): 'INR'},
        "exchangeRate": {('2022-04-05', 0): 1.08, ('2022-04-05', 1): 0.012}
    }

    mocker.patch.object(GsSession.current, '_post', return_value=results)

    response = model.get_currency_exchange_rate(start_date=dt.date(2022, 4, 5),
                                                end_date=dt.date(2022, 4, 5),
                                                format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)
    assert response == currency_exchange_rate_response

    # fitler risk free rates by currency
    expected_filtered_rates = {
        "currency": {('2022-04-05', 1): 'INR'},
        "exchangeRate": {('2022-04-05', 1): 0.012}
    }
    actual_filtered_rates = model.get_currency_exchange_rate(start_date=dt.date(2022, 4, 5),
                                                             end_date=dt.date(2022, 4, 5),
                                                             currencies=[Currency.INR],
                                                             format=ReturnFormat.JSON)

    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)

    assert actual_filtered_rates == expected_filtered_rates

    # test DataFrame return format

    expected_data_frame = get_optional_data_as_dataframe([
        {
            "date": "2022-04-05",
            "currencyRatesData": {
                "exchangeRate": currency_exchange_rate,
                "currency": currencies,
            },
        }
    ], "currencyRatesData")

    expected_data_frame = expected_data_frame.loc[expected_data_frame['currency'].isin(['INR'])]

    actual_data_frame = model.get_currency_exchange_rate(start_date=dt.date(2022, 4, 5),
                                                         end_date=dt.date(2022, 4, 5),
                                                         currencies=[Currency.INR])
    GsSession.current._post.assert_called_with('/risk/models/data/{id}/query'.format(id='model_id'),
                                               query, timeout=200)

    assert_frame_equal(expected_data_frame, actual_data_frame, check_like=True)


if __name__ == "__main__":
    pytest.main([__file__])



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt

import pytest

from gs_quant.api.gs.data import GsDataApi, QueryType
from gs_quant.data import DataContext


def test_build_market_data_query():
    asset_ids = ['MA4B66MW5E27U8P3295']
    query_type = QueryType.IMPLIED_VOLATILITY
    where = {'tenor': '1y', 'strikeReference': 'delta', 'relativeStrike': 0.1}
    source = None
    real_time = False
    measure = "Curve"
    with DataContext(start=dt.date(2013, 1, 1), end=dt.date(2020, 2, 1)):
        queries = GsDataApi.build_market_data_query(asset_ids, query_type, where, source, real_time, measure,
                                                    parallelize_queries=True)
        payload = {'entityIds': ['MA4B66MW5E27U8P3295'],
                   'queryType': 'Implied Volatility',
                   'where': {'tenor': '1y', 'strikeReference': 'delta', 'relativeStrike': 0.1},
                   'source': 'any',
                   'frequency': 'End Of Day',
                   'measures': ['Curve']}

        dates = [{'startDate': dt.date(2013, 1, 1), 'endDate': dt.date(2014, 1, 1)},
                 {'startDate': dt.date(2014, 1, 1), 'endDate': dt.date(2015, 1, 1)},
                 {'startDate': dt.date(2015, 1, 1), 'endDate': dt.date(2016, 1, 1)},
                 {'startDate': dt.date(2016, 1, 1), 'endDate': dt.date(2016, 12, 31)},
                 {'startDate': dt.date(2016, 12, 31), 'endDate': dt.date(2017, 12, 31)},
                 {'startDate': dt.date(2017, 12, 31), 'endDate': dt.date(2018, 12, 31)},
                 {'startDate': dt.date(2018, 12, 31), 'endDate': dt.date(2019, 12, 31)},
                 {'startDate': dt.date(2019, 12, 31), 'endDate': dt.date(2020, 2, 1)}]

        expected = [{'queries': [{**payload, **date_range}]} for date_range in dates]
        assert expected == queries

    with DataContext(start=dt.date(2013, 1, 1), end=dt.date(2025, 1, 1)):
        queries = GsDataApi.build_market_data_query(asset_ids, query_type, where, source, real_time, measure)
        payload = {'entityIds': ['MA4B66MW5E27U8P3295'],
                   'queryType': 'Implied Volatility',
                   'where': {'tenor': '1y', 'strikeReference': 'delta', 'relativeStrike': 0.1},
                   'source': 'any',
                   'frequency': 'End Of Day',
                   'measures': ['Curve'],
                   'startDate': dt.date(2013, 1, 1),
                   'endDate': dt.date(2025, 1, 1)}

        expected = {'queries': [payload]}
        assert expected == queries


if __name__ == '__main__':
    pytest.main(args=["test_queries.py"])



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from dataclasses import dataclass, field
from typing import Tuple, Optional, Union, Iterable

from dataclasses_json import dataclass_json, config

from gs_quant.backtests.backtest_objects import CashAccrualModel
from gs_quant.backtests.backtest_utils import make_list
from gs_quant.backtests.triggers import Trigger
from gs_quant.base import Priceable
from gs_quant.json_convertors import decode_named_instrument, encode_named_instrument, dc_decode


def _backtest_engines():
    from gs_quant.backtests.equity_vol_engine import EquityVolEngine
    from gs_quant.backtests.generic_engine import GenericEngine
    from gs_quant.backtests.predefined_asset_engine import PredefinedAssetEngine
    return [GenericEngine(), PredefinedAssetEngine(), EquityVolEngine()]


@dataclass_json
@dataclass
class Strategy:
    """
    A strategy object on which one may run a backtest
    """
    initial_portfolio: Optional[Union[Tuple[Priceable, ...], dict]] = field(default=None, metadata=config(
        decoder=decode_named_instrument,
        encoder=encode_named_instrument))
    triggers: Union[Trigger, Iterable[Trigger]] = field(default=None,
                                                        metadata=config(decoder=dc_decode(*Trigger.sub_classes(),
                                                                                          allow_missing=True)))
    cash_accrual: CashAccrualModel = None
    risks = None

    def __post_init__(self):
        if not isinstance(self.initial_portfolio, dict):
            self.initial_portfolio = make_list(self.initial_portfolio)
        self.triggers = make_list(self.triggers)
        self.risks = self.get_risks()

    def get_risks(self):
        risk_list = []
        for t in self.triggers:
            risk_list += t.risks if t.risks is not None else []
        return risk_list

    def get_available_engines(self):
        return [engine for engine in _backtest_engines() if engine.supports_strategy(self)]



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import logging
import traceback
from time import sleep
from typing import List, Union, Dict

import deprecation
import numpy as np
import pandas as pd

from gs_quant.api.gs.portfolios import GsPortfolioApi
from gs_quant.api.gs.reports import GsReportApi
from gs_quant.common import Currency, PositionType
from gs_quant.entities.entitlements import Entitlements, EntitlementBlock, User
from gs_quant.entities.entity import PositionedEntity, EntityType, ScenarioCalculationMeasure
from gs_quant.errors import MqError
from gs_quant.errors import MqValueError
from gs_quant.markets.factor import Factor
from gs_quant.markets.portfolio_manager_utils import build_exposure_df, build_portfolio_constituents_df, \
    build_sensitivity_df, get_batched_dates
from gs_quant.markets.report import PerformanceReport, ReportJobFuture
from gs_quant.markets.scenario import FactorScenario
from gs_quant.models.risk_model import MacroRiskModel, ReturnFormat, FactorType, FactorRiskModel
from gs_quant.target.portfolios import RiskAumSource, PortfolioTree
from gs_quant.target.risk_models import RiskModelDataAssetsRequest, RiskModelUniverseIdentifierRequest

_logger = logging.getLogger(__name__)


@deprecation.deprecated(deprecated_in='1.0.10',
                        details='portfolio_manager.CustomAUMDataPoint is now deprecated, please use '
                                'report.CustomAUMDataPoint instead.')
class CustomAUMDataPoint:
    """

    Custom AUM Data Point represents a portfolio's AUM value for a specific date

    """

    def __init__(self,
                 date: dt.date,
                 aum: float):
        self.__date = date
        self.__aum = aum

    @property
    def date(self) -> dt.date:
        return self.__date

    @date.setter
    def date(self, value: dt.date):
        self.__date = value

    @property
    def aum(self) -> float:
        return self.__aum

    @aum.setter
    def aum(self, value: float):
        self.__aum = value


class PortfolioManager(PositionedEntity):
    """
    Portfolio Manager is used to manage Marquee portfolios (setting entitlements, running and retrieving reports, etc)
    """

    def __init__(self,
                 portfolio_id: str):
        """
        Initialize a Portfolio Manager

        :param portfolio_id: Portfolio ID
        """
        self.__portfolio_id = portfolio_id
        PositionedEntity.__init__(self, portfolio_id, EntityType.PORTFOLIO)

    @property
    def portfolio_id(self) -> str:
        return self.__portfolio_id

    @portfolio_id.setter
    def portfolio_id(self, value: str):
        self.__portfolio_id = value

    def get_performance_report(self, tags: Dict = None) -> PerformanceReport:
        """
        Get performance report associated with a portfolio

        :param tags: If the portfolio is a fund of funds, pass in a dictionary corresponding to the tag values
        to retrieve results for a sub-portfolio

        :return: returns the PerformanceReport associated with portfolio if one exists
        """
        reports = GsReportApi.get_reports(limit=100,
                                          position_source_type='Portfolio',
                                          position_source_id=self.id,
                                          report_type='Portfolio Performance Analytics',
                                          tags=tags,
                                          scroll='1m')

        # If tags is set to None, it returns all PPA reports for the portfolio,
        # and returning reports[0] can return any one PPA, so specifically if tags is None it means we are
        # looking for the root PPA, we need to explicitly filter out and get the one report where tags is None
        if tags is None:
            reports = [report for report in reports if report.parameters.tags is None]
        if len(reports) == 0:
            raise MqError('No performance report found.')
        return PerformanceReport.from_target(reports[0])

    def schedule_reports(self,
                         start_date: dt.date = None,
                         end_date: dt.date = None,
                         backcast: bool = False,
                         months_per_batch: int = None):
        """
        Schedule all reports associated with a portfolio. If months_per_batch is passed and reports are historical, then
        schedule them in batches of bathc_period months. Backcasted reported can't be batched.

        :param start_date: start date of report job (optional)
        :param end_date: end date of report job (optional)
        :param backcast: true if reports should be backcasted; defaults to false
        :param months_per_batch: batch size of historical report job schedules in number of months; defaults to false
        if set, historical reports are scheduled in batches of size less than or equal to the given number of months

        **Examples**

        >>> pm = PortfolioManager("PORTFOLIO ID")
        >>> pm.schedule_reports(backcast=False)

        for a portfolio having hisotry greater than 1 yr, use months_per_batch to schedule reports in batches of
        6 months
        >>> pm = PortfolioManager("PORTFOLIO ID")
        >>> pm.schedule_reports(backcast=False, months_per_batch=6)
        """
        if months_per_batch is None or backcast:
            if backcast:
                print('Batching of schedule reports is only supported for historical reports. '
                      'Backcasted reports will be scheduled for the full date range')
            GsPortfolioApi.schedule_reports(self.__portfolio_id, start_date, end_date, backcast=backcast)
        else:
            # Process input
            if months_per_batch <= 0:
                raise MqValueError(f'Invalid input, months_per_batch {months_per_batch} should be greater than 0')

            if end_date is None or start_date is None:
                hints = self.get_schedule_dates(backcast=backcast)
                start_date = hints[0] if start_date is None else start_date
                end_date = hints[1] if end_date is None else end_date

            if start_date >= end_date:
                raise MqValueError(f'Invalid input, start date {start_date} should be before end date {end_date}')

            # Validate input date range against position dates
            position_dates = self.get_position_dates()
            position_dates = [p for p in position_dates if end_date >= p >= start_date]

            if len(position_dates) == 0:
                raise MqValueError('Portfolio does not have any positions in the given date range')

            if start_date not in position_dates:
                # There should be positions on the start date for historical reports
                raise MqError('Cannot schedule historical report because the first set of positions within the '
                              f'scheduling date range need to be on this date {start_date}')
            else:
                # preparation for the first batch
                position_dates.remove(start_date)

            if end_date not in position_dates:
                # make sure to process the last batch
                position_dates.append(end_date)

            # Create batches using sliding window
            batch_boundaries = [start_date]
            prev_date = start_date
            for i, d in enumerate(position_dates):
                current_batch = d - prev_date
                if current_batch.days > months_per_batch * 30 and i > 0:
                    # split the current window at the previous date
                    prev_date = position_dates[i - 1]
                    batch_boundaries.append(prev_date)

            batch_boundaries.append(end_date)

            # Schedule reports in batches
            print(f'Scheduling reports from {start_date} to {end_date} in {len(batch_boundaries) - 1} batches')

            for i in range(len(batch_boundaries) - 1):
                print(f'Scheduling for {batch_boundaries[i]} to {batch_boundaries[i + 1]}')
                GsPortfolioApi.schedule_reports(self.__portfolio_id,
                                                batch_boundaries[i],
                                                batch_boundaries[i + 1],
                                                backcast=backcast)
                if i > 0 and i % 10 == 0:
                    sleep(6)

    def run_reports(self,
                    start_date: dt.date = None,
                    end_date: dt.date = None,
                    backcast: bool = False,
                    is_async: bool = True,
                    months_per_batch: int = None) -> List[Union[pd.DataFrame, ReportJobFuture]]:
        """
        Run all reports associated with a portfolio

        :param start_date: start date of report job
        :param end_date: end date of report job
        :param backcast: true if reports should be backcasted; defaults to false
        :param is_async: true if reports should run asynchronously; defaults to true
        :param months_per_batch: batch size of historical report job schedules in number of months; defaults to false
        if set, historical reports are scheduled in batches of size less than or equal to the given number of months
        :return: if is_async is true, returns a list of ReportJobFuture objects; if is_async is false, returns a list
        of dataframe objects containing report results for all portfolio results

        **Examples**

        >>> pm = PortfolioManager("PORTFOLIO ID")
        >>> report_results = pm.run_reports(backcast=True)

        for longer hisotry, use months_per_batch to schedule reports in batches of 6 months
        >>> pm = PortfolioManager("PORTFOLIO ID")
        >>> pm.schedule_reports(backcast=False, months_per_batch=6)
        """
        self.schedule_reports(start_date, end_date, backcast, months_per_batch)
        reports = self.get_reports()
        report_futures = [report.get_most_recent_job() for report in reports]
        if is_async:
            return report_futures
        counter = 100
        while counter > 0:
            is_done = [future.done() for future in report_futures]
            if False not in is_done:
                return [job_future.result() for job_future in report_futures]
            sleep(6)
        raise MqValueError(f'Your reports for Portfolio {self.__portfolio_id} are taking longer than expected '
                           f'to finish. Please contact the Marquee Analytics team at '
                           f'gs-marquee-analytics-support@gs.com')

    def set_entitlements(self,
                         entitlements: Entitlements):
        """
        Set the entitlements of a portfolio

        :param entitlements: Entitlements object

        **Examples**

        >>> portfolio_admin_emails = ['LIST OF ADMIN EMAILS']
        >>> portfolio_viewer_emails = ['LIST OF VIEWER EMAILS']
        >>> admin_entitlements = EntitlementBlock(users=User.get_many(emails=portfolio_admin_emails))
        >>> view_entitlements = EntitlementBlock(users=User.get_many(emails=portfolio_viewer_emails))
        >>>
        >>> entitlements = Entitlements(
        >>>    view=view_entitlements,
        >>>    admin=admin_entitlements
        >>> )
        >>>
        >>> pm = PortfolioManager("PORTFOLIO ID")
        >>> pm.set_entitlements(entitlements)

        **See Also**
        :func: PortfolioManager.share
        :func: PortfolioManager.get_entitlements
        """
        entitlements_as_target = entitlements.to_target()
        portfolio_as_target = GsPortfolioApi.get_portfolio(self.__portfolio_id)
        portfolio_as_target.entitlements = entitlements_as_target
        GsPortfolioApi.update_portfolio(portfolio_as_target)

    def share(self, emails: List[str], admin: bool = False):
        """
        Share a portfolio with a list of emails

        :param emails: list of emails to share the portfolio with
        :param admin: true if the users should have admin access; defaults to false

        **Examples**

        >>> pm = PortfolioManager("PORTFOLIO ID")
        >>> pm.share(['EMAIL1', 'EMAIL2'])
        """
        current_entitlements = self.get_entitlements()
        users = User.get_many(emails=emails)
        found_emails = [user.email for user in users]
        if len(found_emails) != len(emails):
            missing_emails = [email for email in emails if email not in found_emails]
            raise MqValueError(f'Users with emails {missing_emails} not found')
        if admin:
            current_entitlements.admin = EntitlementBlock(users=set(current_entitlements.admin.users + users))
        current_entitlements.view = EntitlementBlock(users=set(current_entitlements.view.users + users))
        self.set_entitlements(current_entitlements)

    def set_currency(self, currency: Currency):
        """
        Set the currency of a portfolio

        :param currency: Currency
        """
        portfolio_as_target = GsPortfolioApi.get_portfolio(self.__portfolio_id)
        portfolio_as_target.currency = currency
        GsPortfolioApi.update_portfolio(portfolio_as_target)

    def get_tag_name_hierarchy(self) -> List:
        """
        Get the list of tags by name by which a portfolio's fund of funds are structured in that order

        :return: a list of tag names
        """
        portfolio = GsPortfolioApi.get_portfolio(self.portfolio_id)
        return list(portfolio.tag_name_hierarchy) if portfolio.tag_name_hierarchy else None

    def set_tag_name_hierarchy(self, tag_names: List):
        """
        Set the list of tags by name by which a portfolio's fund of funds are structured in that order

        :param tag_names: a list of tag names in order of the new fund of funds structure
        """
        portfolio = GsPortfolioApi.get_portfolio(self.portfolio_id)
        portfolio.tag_name_hierarchy = tag_names
        GsPortfolioApi.update_portfolio(portfolio)

    def update_portfolio_tree(self):
        """
        After a modification is made on your portfolio (reports are added/modified, the tag name hierarchy is changed,
        etc), run this function so those changes reflect across all the portfolio's sub-portfolios
        """
        GsPortfolioApi.update_portfolio_tree(self.portfolio_id)

    def get_portfolio_tree(self) -> PortfolioTree:
        """
        Get the portfolio tree of a fund of funds portfolio
        """
        return GsPortfolioApi.get_portfolio_tree(self.portfolio_id)

    def get_all_fund_of_fund_tags(self) -> List:
        """
        If the portfolio is a fund of funds, this function will retrieve a list of dictionaries of all the tag sets
        associated with the sub-portfolios in the portfolio.

        :return: a list of tags as dictionaries
        """
        tag_dicts = []
        for r in self.get_reports():
            if r.parameters.tags is not None:
                tags_as_dict = {tag.name: tag.value for tag in r.parameters.tags}
                if tags_as_dict not in tag_dicts:
                    tag_dicts.append(tags_as_dict)
        tag_dicts.sort(key=lambda dictionary: len(dictionary.keys()))
        return tag_dicts

    def get_schedule_dates(self,
                           backcast: bool = False) -> List[dt.date]:
        """
        Get recommended start and end dates for a portfolio report scheduling job

        :param backcast: true if reports should be backcasted
        :return: a list of two dates, the first is the suggested start date and the second is the suggested end date
        """
        return GsPortfolioApi.get_schedule_dates(self.id, backcast)

    @deprecation.deprecated(deprecated_in='1.0.10',
                            details='PortfolioManager.get_aum_source is now deprecated, please use '
                                    'PerformanceReport.get_aum_source instead.')
    def get_aum_source(self) -> RiskAumSource:
        """
        Get portfolio AUM Source

        :return: aum source
        """
        portfolio = GsPortfolioApi.get_portfolio(self.portfolio_id)
        return portfolio.aum_source if portfolio.aum_source is not None else RiskAumSource.Long

    @deprecation.deprecated(deprecated_in='1.0.10',
                            details='PortfolioManager.set_aum_source is now deprecated, please use '
                                    'PerformanceReport.set_aum_source instead.')
    def set_aum_source(self,
                       aum_source: RiskAumSource):
        """
        Set portfolio AUM Source

        :param aum_source: aum source for portfolio
        :return: aum source
        """
        portfolio = GsPortfolioApi.get_portfolio(self.portfolio_id)
        portfolio.aum_source = aum_source
        GsPortfolioApi.update_portfolio(portfolio)

    @deprecation.deprecated(deprecated_in='1.0.10',
                            details='PortfolioManager.get_custom_aum is now deprecated, please use '
                                    'PerformanceReport.get_custom_aum instead.')
    def get_custom_aum(self,
                       start_date: dt.date = None,
                       end_date: dt.date = None) -> List[CustomAUMDataPoint]:
        """
        Get AUM data for portfolio

        :param start_date: start date
        :param end_date: end date
        :return: list of AUM data between the specified range
        """
        aum_data = GsPortfolioApi.get_custom_aum(self.portfolio_id, start_date, end_date)
        return [CustomAUMDataPoint(date=dt.datetime.strptime(data['date'], '%Y-%m-%d'),
                                   aum=data['aum']) for data in aum_data]

    @deprecation.deprecated(deprecated_in='1.0.10',
                            details='PortfolioManager.get_aum is now deprecated, please use '
                                    'PerformanceReport.get_aum instead.')
    def get_aum(self,
                start_date: dt.date,
                end_date: dt.date):
        """
        Get AUM data for portfolio

        :param start_date: start date
        :param end_date: end date
        :return: dictionary of dates with corresponding AUM values
        """
        aum_source = self.get_aum_source()
        if aum_source == RiskAumSource.Custom_AUM:
            aum = self.get_custom_aum(start_date=start_date, end_date=end_date)
            return {aum_point.date.strftime('%Y-%m-%d'): aum_point.aum for aum_point in aum}
        if aum_source == RiskAumSource.Long:
            aum = self.get_performance_report().get_long_exposure(start_date=start_date, end_date=end_date)
            return {row['date']: row['longExposure'] for index, row in aum.iterrows()}
        if aum_source == RiskAumSource.Short:
            aum = self.get_performance_report().get_short_exposure(start_date=start_date, end_date=end_date)
            return {row['date']: row['shortExposure'] for index, row in aum.iterrows()}
        if aum_source == RiskAumSource.Gross:
            aum = self.get_performance_report().get_gross_exposure(start_date=start_date, end_date=end_date)
            return {row['date']: row['grossExposure'] for index, row in aum.iterrows()}
        if aum_source == RiskAumSource.Net:
            aum = self.get_performance_report().get_net_exposure(start_date=start_date, end_date=end_date)
            return {row['date']: row['netExposure'] for index, row in aum.iterrows()}

    @deprecation.deprecated(deprecated_in='1.0.10',
                            details='PortfolioManager.upload_custom_aum is now deprecated, please use '
                                    'PerformanceReport.upload_custom_aum instead.')
    def upload_custom_aum(self,
                          aum_data: List[CustomAUMDataPoint],
                          clear_existing_data: bool = None):
        """
        Add AUM data for portfolio

        :param aum_data: list of AUM data to upload
        :param clear_existing_data: delete all previously uploaded AUM data for the portfolio (defaults to false)
        """
        formatted_aum_data = [{'date': data.date.strftime('%Y-%m-%d'), 'aum': data.aum} for data in aum_data]
        GsPortfolioApi.upload_custom_aum(self.portfolio_id, formatted_aum_data, clear_existing_data)

    @deprecation.deprecated(deprecated_in="0.9.110",
                            details="Please use the get_pnl_contribution on your portfolio's performance report using"
                                    "the PerformanceReport class")
    def get_pnl_contribution(self,
                             start_date: dt.date = None,
                             end_date: dt.date = None,
                             currency: Currency = None,
                             tags: Dict = None) -> pd.DataFrame:
        """
        Get PnL Contribution of your portfolio broken down by constituents

        :param start_date: optional start date
        :param end_date: optional end date
        :param currency: optional currency; defaults to your portfolio's currency
        :param tags: If the portfolio is a fund of funds, pass in a dictionary corresponding to the tag values
        to retrieve results for a sub-portfolio
        :return: a Pandas DataFrame of results
        """
        performance_report_id = None if tags is None else self.get_performance_report(tags).id
        return pd.DataFrame(GsPortfolioApi.get_attribution(self.portfolio_id, start_date, end_date,
                                                           currency, performance_report_id))

    def get_macro_exposure(self,
                           model: MacroRiskModel,
                           date: dt.date,
                           factor_type: FactorType,
                           factor_categories: List[Factor] = [],
                           get_factors_by_name: bool = True,
                           tags: Dict = None,
                           return_format: ReturnFormat = ReturnFormat.DATA_FRAME
                           ) -> Union[Dict, pd.DataFrame]:
        """
        Get portfolio and asset exposure to macro factors or macro factor categories

        :param model: the macro risk model
        :param date: date for which to get exposure
        :param factor_type: whether to get exposure to factor categories or factors.
        :param factor_categories: Get portfolio exposure to these factor categories. Must be valid Factor Categories.
        If factor_type is Factor, get exposure to factors that are grouped in these factor categories.
        If empty, return exposure to all factor categories/factors.
        :param get_factors_by_name: whether to identify factors by their name or identifier
        :param tags: If the portfolio is a fund of funds, pass in a dictionary corresponding to the tag values
        to retrieve results for a sub-portfolio
        :param return_format: whether to return a dict or a pandas dataframe

        :return: a Pandas Dataframe or a Dict of portfolio exposure to macro factors

        **Examples**

        >>> model = MacroRiskModel.get(\"MODEL\")
        >>> pm = PortfolioManager("PORTFOLIO ID")
        >>> exposure_dataframe = pm.get_macro_exposure(
        >>>     model=model,
        >>>     date=dt.date(2022, 1, 1),
        >>>     factor_type=FactorType.Factor
        >>> ).sort_values(
        >>>     by=[\"Total Factor Exposure\"],
        >>>     axis=1,
        >>>     ascending=False
        >>> )
        """
        performance_report = self.get_performance_report(tags)

        # Get portfolio constituents
        constituents_and_notional_df = build_portfolio_constituents_df(performance_report, date). \
            rename(columns={"name": "Asset Name", "netExposure": "Notional"})

        # Query universe sensitivity
        universe = constituents_and_notional_df.index.dropna().tolist()
        universe_sensitivities_df = build_sensitivity_df(universe, model, date, factor_type, get_factors_by_name)

        # Remove assets without exposure
        assets_with_exposure = list(universe_sensitivities_df.index.values)
        if not assets_with_exposure:
            logging.warning("The Portfolio is not exposed to any of the requested macro factors")
            return pd.DataFrame()

        constituents_and_notional_df = constituents_and_notional_df.loc[assets_with_exposure]

        factor_data = model.get_factor_data(date, date, factor_type=FactorType.Factor) \
            if factor_type == FactorType.Factor else pd.DataFrame()
        exposure_df = build_exposure_df(constituents_and_notional_df, universe_sensitivities_df,
                                        factor_categories, factor_data, get_factors_by_name)

        if return_format == ReturnFormat.JSON:
            return exposure_df.to_dict()

        return exposure_df

    def get_factor_scenario_analytics(self,
                                      scenarios: List[FactorScenario],
                                      date: dt.date,
                                      measures: List[ScenarioCalculationMeasure],
                                      risk_model: str = None,
                                      return_format: ReturnFormat = ReturnFormat.DATA_FRAME) -> \
            Union[Dict, Union[Dict, pd.DataFrame]]:
        """Given a list of factor scenarios (historical simulation and/or custom shocks), return the estimated pnl
         of the given positioned entity.
         :param scenarios: List of factor-based scenarios
         :param date: date to run scenarios.
         :param measures: which metrics to return
         :param risk_model: valid risk model ID
         :param return_format: whether to return data formatted in a dataframe or as a dict

          **Examples**

            >>> from gs_quant.session import GsSession, Environment
            >>> from gs_quant.markets.portfolio_manager import PortfolioManager, ReturnFormat
            >>> from gs_quant.entities.entity import ScenarioCalculationMeasure, PositionedEntity
            >>> from gs_quant.markets.scenario import Scenario

          Get scenarios
            >>> covid_19_omicron = Scenario.get_by_name("Covid 19 Omicron (v2)") # historical simulation
            >>> custom_shock = Scenario.get_by_name("Shocking factor by x% (Propagated)") # custom shock
            >>> risk_model = "RISK_MODEL_ID" # valid risk model ID

          Instantiate your positionedEntity. Here, we are using one of its subclasses, PortfolioManager

            >>> pm = PortfolioManager(portfolio_id="PORTFOLIO_ID")

          Set the date you wish to run your scenario on

           >>> date = dt.date(2023, 3, 7)

          Run scenario and get estimated impact on your positioned entity

          >>> scenario_analytics = pm.get_factor_scenario_analytics(
          ...     scenarios=[covid_19_omicron, beta_propagated],
          ...     date=date,
          ...     measures=[ScenarioCalculationMeasure.SUMMARY,
          ...               ScenarioCalculationMeasure.ESTIMATED_FACTOR_PNL,
          ...               ScenarioCalculationMeasure.ESTIMATED_PNL_BY_SECTOR,
          ...               ScenarioCalculationMeasure.ESTIMATED_PNL_BY_REGION,
          ...               ScenarioCalculationMeasure.ESTIMATED_PNL_BY_DIRECTION,
          ...               ScenarioCalculationMeasure.ESTIMATED_PNL_BY_ASSET],
          ...     risk_model=risk_model)

          By default, the result will be returned in a dict with keys as the measures/metrics requested and values as
          the scenario calculation results formatted in a dataframe. To get the results in a dict, specify the return
          format as JSON
         """

        return super().get_factor_scenario_analytics(scenarios, date, measures, risk_model, return_format)

    def get_risk_model_predicted_beta(self, start_date: dt.date, end_date: dt.date, risk_model_id: str,
                                      default_beta_value: int = 1, tags: Dict = None) -> pd.DataFrame:
        """
        Get the predicted beta of a portfolio as estimated by the provided risk model.
        The beta is calculated using the factor risk model data. The function retrieves the factor risk report
        associated with the portfolio and the risk model ID, and then calculates the predicted beta for each date
        in the specified range.

        :param start_date: Start date for the beta calculation
        :param end_date: End date for the beta calculation
        :param risk_model_id: Risk model ID to be used for the calculation
        :param default_beta_value: Default beta value to be used if no data is available for a specific date
        :param tags: If the portfolio is a fund of funds, pass in a dictionary corresponding to the tag values
        to retrieve results for a sub-portfolio

        :return: A DataFrame containing the predicted beta for each date in the specified range

        **Examples**
        >>> pm = PortfolioManager("PORTFOLIO ID")
        >>> beta_df = pm.get_risk_model_predicted_beta(start_date=dt.date(2023, 1, 1),
        ... end_date=dt.date(2023, 12, 31), risk_model_id="RISK_MODEL_ID")
        >>> print(beta_df)
        >>> # Output:

        date        beta
        2023-01-01  1.05
        2023-01-02  1.10
        2023-01-03  1.02
        """
        performance_report = self.get_performance_report(tags=tags)
        risk_model = FactorRiskModel.get(risk_model_id)
        risk_model_dates: List[dt.date] = risk_model.get_dates(start_date=start_date, end_date=end_date)
        batched_dates = get_batched_dates(risk_model_dates, batch_size=10)
        risk_model_predicted_beta_timeseries = pd.DataFrame()
        asset_level_betas = pd.DataFrame()
        portfolio_position_net_weights = pd.DataFrame()
        for batch in batched_dates:
            start_date, end_date = batch[0], batch[-1]
            try:
                date_wise_net_weight = performance_report.get_position_net_weights(start_date=start_date,
                                                                                   end_date=end_date,
                                                                                   asset_metadata_fields=["gsid"],
                                                                                   include_all_business_days=True,
                                                                                   position_type=PositionType.CLOSE)
                date_wise_net_weight = date_wise_net_weight.dropna().pivot_table(index='positionDate', columns='gsid',
                                                                                 values='netWeight', aggfunc='sum')
                portfolio_position_gsids = [gsid for gsid in date_wise_net_weight.columns.tolist() if gsid is not None]
                asset_level_betas_batch = (risk_model
                                           .get_predicted_beta(start_date=start_date, end_date=end_date,
                                                               assets=RiskModelDataAssetsRequest(
                                                                   identifier=RiskModelUniverseIdentifierRequest.gsid,
                                                                   universe=tuple(portfolio_position_gsids))))
                if asset_level_betas_batch.isna().all().all():
                    logging.warning(
                        f"Risk model predicted beta not available for risk model {risk_model_id} for dates: {batch}")
                    continue
                else:
                    asset_level_betas = pd.concat([asset_level_betas, asset_level_betas_batch], axis=0)
                    portfolio_position_net_weights = pd.concat([portfolio_position_net_weights,
                                                                date_wise_net_weight], axis=0)
            except Exception as ex:
                raise MqError(
                    f"Risk model predicted beta cannot be calculated for risk model {risk_model_id} "
                    f"from {start_date} to {end_date} because of exception: {ex}, traceback: {traceback.format_exc()}")
        if asset_level_betas.empty:
            return pd.DataFrame()
        asset_level_betas = asset_level_betas.fillna(default_beta_value)
        risk_model_predicted_beta_timeseries = asset_level_betas * portfolio_position_net_weights
        risk_model_predicted_beta_timeseries = (risk_model_predicted_beta_timeseries.sum(axis=1).rename('beta').
                                                replace({0: np.nan}).ffill())
        return risk_model_predicted_beta_timeseries.reset_index().rename(columns={'index': 'date', 0: 'beta'})



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import json
from unittest.mock import patch

import numpy as np
import pandas as pd

from gs_quant.backtests.actions import AddTradeAction, HedgeAction, ExitTradeAction, AddScaledTradeAction, \
    ScalingActionType, ExitAllPositionsAction
from gs_quant.backtests.backtest_objects import (ScaledTransactionModel, AggregateTransactionModel, TransactionAggType,
                                                 ConstantTransactionModel, BackTest)
from gs_quant.backtests.data_sources import GenericDataSource, MissingDataStrategy, GsDataSource
from gs_quant.backtests.generic_engine import GenericEngine
from gs_quant.backtests.strategy import Strategy
from gs_quant.backtests.triggers import TriggerDirection, DateTrigger, DateTriggerRequirements, NotTrigger, \
    RiskTriggerRequirements, \
    StrategyRiskTrigger, PeriodicTriggerRequirements, PeriodicTrigger, MktTrigger, MktTriggerRequirements, \
    IntradayPeriodicTrigger, IntradayTriggerRequirements, AggregateTrigger, AggregateTriggerRequirements, \
    PortfolioTrigger, PortfolioTriggerRequirements, MeanReversionTrigger, MeanReversionTriggerRequirements, \
    NotTriggerRequirements
from gs_quant.common import Currency, PayReceive, OptionType, OptionStyle
from gs_quant.instrument import FXOption, FXForward, IRSwaption, IRSwap, EqOption
from gs_quant.json_encoder import JSONEncoder
from gs_quant.markets import PricingContext
from gs_quant.markets.portfolio import Portfolio
from gs_quant.risk import Price, FXDelta, DollarPrice, IRDelta
from gs_quant.target.measures import EqDelta, EqVega
from gs_quant.test.utils.mock_calc import MockCalc


def mock_pricing_context(self):
    context = PricingContext(set_parameters_only=True, is_batch=False, show_progress=True)
    return context


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_generic_engine_simple(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 1)
        # end_date = dt.date(2021, 12, 3)

        # Define trade
        call = FXOption(buy_sell='Buy', option_type='Call', pair='USDJPY', strike_price='ATMF',
                        notional_amount=1e5, expiration_date='2y', name='2y_call')

        # Periodic trigger: based on frequency
        freq = '1m'

        # trig_req = PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency=freq)
        trig_req = DateTriggerRequirements(dates=[start_date])
        actions = AddTradeAction(call, freq, name='Action1')

        # starting with empty portfolio (first arg to Strategy), apply actions on trig_req
        triggers = DateTrigger(trig_req, actions)

        strategy = Strategy(None, triggers)

        # run backtest daily
        engine = GenericEngine()
        backtest = engine.run_backtest(strategy,
                                       states=[dt.date(2021, 12, 1), dt.date(2021, 12, 2), dt.date(2021, 12, 3)],
                                       show_progress=True)
        summary = backtest.result_summary
        assert len(summary) == 3
        assert round(summary[Price].sum()) == 2424
        assert round(summary['Cumulative Cash'][-1]) == 0


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_hedge_action_risk_trigger(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 1)
        # end_date = dt.date(2021, 12, 3)

        # Define trade
        call = FXOption(buy_sell='Buy', option_type='Call', pair='USDJPY', strike_price='ATMF',
                        notional_amount=1e5, expiration_date='2y', name='2y_call')

        hedge_risk = FXDelta(aggregation_level='Type')
        fwd_hedge = FXForward(pair='USDJPY', settlement_date='2y', notional_amount=1e5, name='2y_forward')

        trig_req = RiskTriggerRequirements(risk=hedge_risk, trigger_level=0, direction=TriggerDirection.ABOVE)
        action_hedge = HedgeAction(hedge_risk, fwd_hedge, '2b', name='HedgeAction1')

        triggers = StrategyRiskTrigger(trig_req, action_hedge)

        with PricingContext(pricing_date=start_date, use_historical_diddles_only=True):
            fut = call.resolve(in_place=False)

        call = fut.result()

        strategy = Strategy(call, triggers)

        # run backtest daily
        engine = GenericEngine()
        # backtest = engine.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)
        backtest = engine.run_backtest(strategy,
                                       states=[dt.date(2021, 12, 1), dt.date(2021, 12, 2), dt.date(2021, 12, 3)],
                                       show_progress=True)
        summary = backtest.result_summary
        assert len(summary) == 3
        assert round(summary[hedge_risk].sum()) == 0
        assert round(summary['Cumulative Cash'][-1]) == -7090
        assert Price in summary.columns


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_hedge_without_risk(mocker):
    with MockCalc(mocker):
        # Define trade
        call = FXOption(buy_sell='Buy', option_type='Call', pair='USDJPY', strike_price='ATMF',
                        notional_amount=1e5, expiration_date='2y', name='2y_call')

        trig_req = PeriodicTriggerRequirements(start_date=dt.date(2021, 12, 1), end_date=dt.date(2021, 12, 3),
                                               frequency='1b')

        action = AddTradeAction(call, '1b', name='AddAction1')

        hedge_risk = FXDelta(aggregation_level='Type')
        fwd_hedge = FXForward(pair='USDJPY', settlement_date='2y', notional_amount=1e5, name='2y_forward')

        hedge_trig_req = PeriodicTriggerRequirements(start_date=dt.date(2021, 11, 1), end_date=dt.date(2022, 1, 1),
                                                     frequency='1b')
        action_hedge = HedgeAction(hedge_risk, fwd_hedge, '2b', name='HedgeAction1')

        triggers = [PeriodicTrigger(trig_req, action), PeriodicTrigger(hedge_trig_req, action_hedge)]

        strategy = Strategy(None, triggers)

        # run backtest daily
        engine = GenericEngine()
        # backtest = engine.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)
        backtest = engine.run_backtest(strategy,
                                       states=[dt.date(2021, 12, 1), dt.date(2021, 12, 2), dt.date(2021, 12, 3)],
                                       show_progress=True)
        summary = backtest.result_summary
        assert len(summary) == 3
        assert round(summary[hedge_risk].sum()) == 0
        assert round(summary['Cumulative Cash'][-1]) == -6579
        assert Price in summary.columns


s = pd.Series({dt.date(2021, 10, 1): 0.984274,
               dt.date(2021, 10, 4): 1.000706,
               dt.date(2021, 10, 5): 1.044055,
               dt.date(2021, 10, 6): 1.095361,
               dt.date(2021, 10, 7): 1.129336,
               dt.date(2021, 10, 8): 1.182954,
               dt.date(2021, 10, 12): 1.200108,
               dt.date(2021, 10, 13): 1.220607,
               dt.date(2021, 10, 14): 1.172837,
               dt.date(2021, 10, 15): 1.163660,
               dt.date(2021, 10, 18): 1.061084,
               dt.date(2021, 10, 19): 1.025012,
               dt.date(2021, 10, 20): 1.018035,
               dt.date(2021, 10, 21): 1.080751,
               dt.date(2021, 10, 22): 1.069340,
               dt.date(2021, 10, 25): 1.033413})


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_mkt_trigger_data_sources(mocker):
    with MockCalc(mocker):
        action = AddTradeAction(IRSwaption(notional_currency='USD', expiration_date='1y', termination_date='1y'),
                                'expiration_date', name='Action1')
        data_source = GenericDataSource(s, MissingDataStrategy.fill_forward)
        mkt_trigger = MktTrigger(MktTriggerRequirements(data_source, 1.1, TriggerDirection.ABOVE), action)
        strategy = Strategy(None, mkt_trigger)

        engine = GenericEngine()

        backtest = engine.run_backtest(strategy, states=list(s.index), show_progress=True)

        summary = backtest.result_summary
        ledger = backtest.trade_ledger()
        assert len(summary) == 12
        assert len(ledger) == 6
        assert round(summary[Price].sum()) == 25163614
        assert round(summary['Cumulative Cash'][-1]) == -2153015


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_exit_action_noarg(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 6)
        end_date = dt.date(2021, 12, 10)

        # Define trade
        irswap = IRSwap(PayReceive.Receive, '10y', Currency.USD, notional_amount=1e5, name='swap')

        trig_req_add = PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='1b')
        trig_req_exit = PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='2b')
        actions_add = AddTradeAction(irswap, name='Action1')
        actions_exit = ExitTradeAction(name='ExitAction1')

        triggers = [PeriodicTrigger(trig_req_add, actions_add), PeriodicTrigger(trig_req_exit, actions_exit)]
        strategy = Strategy(None, triggers)

        # run backtest daily
        engine = GenericEngine()
        # backtest = engine.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)
        backtest = engine.run_backtest(strategy,
                                       states=[dt.date(2021, 12, 6), dt.date(2021, 12, 7), dt.date(2021, 12, 8),
                                               dt.date(2021, 12, 9), dt.date(2021, 12, 10)], end=end_date,
                                       show_progress=True)

        trade_ledger = backtest.trade_ledger().to_dict('index')

        assert trade_ledger['Action1_swap_2021-12-06']['Open'] == dt.date(2021, 12, 6)
        assert trade_ledger['Action1_swap_2021-12-06']['Close'] == dt.date(2021, 12, 6)
        assert trade_ledger['Action1_swap_2021-12-07']['Open'] == dt.date(2021, 12, 7)
        assert trade_ledger['Action1_swap_2021-12-07']['Close'] == dt.date(2021, 12, 8)


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_exit_action_emptyresults(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 6)
        end_date = dt.date(2021, 12, 10)

        # Define trade
        irswap = IRSwap(PayReceive.Receive, '10y', Currency.USD, notional_amount=1e5, name='swap')

        trig_req_add = PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='2b')
        trig_req_exit = PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='1b')
        actions_add = AddTradeAction(irswap, name='Action1')
        actions_exit = ExitTradeAction(name='ExitAction1')

        triggers = [PeriodicTrigger(trig_req_add, actions_add), PeriodicTrigger(trig_req_exit, actions_exit)]
        strategy = Strategy(None, triggers)

        # run backtest daily
        engine = GenericEngine()
        # backtest = engine.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)
        backtest = engine.run_backtest(strategy,
                                       states=[dt.date(2021, 12, 6), dt.date(2021, 12, 7), dt.date(2021, 12, 8),
                                               dt.date(2021, 12, 9), dt.date(2021, 12, 10)], end=end_date,
                                       show_progress=True)

        trade_ledger = backtest.trade_ledger().to_dict('index')

        assert trade_ledger['Action1_swap_2021-12-06']['Open'] == dt.date(2021, 12, 6)
        assert trade_ledger['Action1_swap_2021-12-06']['Close'] == dt.date(2021, 12, 6)
        assert trade_ledger['Action1_swap_2021-12-08']['Open'] == dt.date(2021, 12, 8)
        assert trade_ledger['Action1_swap_2021-12-08']['Close'] == dt.date(2021, 12, 8)
        assert trade_ledger['Action1_swap_2021-12-10']['Open'] == dt.date(2021, 12, 10)
        assert trade_ledger['Action1_swap_2021-12-10']['Close'] == dt.date(2021, 12, 10)


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_exit_action_bytradename(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 6)
        end_date = dt.date(2021, 12, 10)

        # Define trade
        irswap1 = IRSwap(PayReceive.Receive, '10y', Currency.USD, notional_amount=1e5, name='swap1')
        irswap2 = IRSwap(PayReceive.Pay, '5y', Currency.USD, notional_amount=1e5, name='swap2')

        trig_req_add = PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='1b')
        trig_req_exit = PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='2b')
        actions_add = AddTradeAction([irswap1, irswap2], name='Action1')
        actions_exit = ExitTradeAction('swap1', name='ExitAction1')

        triggers = [PeriodicTrigger(trig_req_add, actions_add), PeriodicTrigger(trig_req_exit, actions_exit)]
        strategy = Strategy(None, triggers)

        # run backtest daily
        engine = GenericEngine()
        # backtest = engine.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)
        backtest = engine.run_backtest(strategy,
                                       states=[dt.date(2021, 12, 6), dt.date(2021, 12, 7), dt.date(2021, 12, 8),
                                               dt.date(2021, 12, 9), dt.date(2021, 12, 10)], end=end_date,
                                       show_progress=True)

        trade_ledger = backtest.trade_ledger().to_dict('index')

        assert trade_ledger['Action1_swap1_2021-12-06']['Open'] == dt.date(2021, 12, 6)
        assert trade_ledger['Action1_swap1_2021-12-06']['Close'] == dt.date(2021, 12, 6)
        assert trade_ledger['Action1_swap1_2021-12-07']['Open'] == dt.date(2021, 12, 7)
        assert trade_ledger['Action1_swap1_2021-12-07']['Close'] == dt.date(2021, 12, 8)
        assert trade_ledger['Action1_swap2_2021-12-06']['Status'] == 'open'
        assert trade_ledger['Action1_swap2_2021-12-07']['Status'] == 'open'
        assert trade_ledger['Action1_swap2_2021-12-10']['Status'] == 'open'


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_add_scaled_action(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 6)
        end_date = dt.date(2021, 12, 10)

        scale_factor = 7

        # Define instruments for strategy
        # Portfolio of two eq options
        call = EqOption('.STOXX50E', expiration_date='1m', strike_price='ATM', option_type=OptionType.Call,
                        option_style=OptionStyle.European, name='call')
        put = EqOption('.STOXX50E', expiration_date='1m', strike_price='ATM', option_type=OptionType.Put,
                       option_style=OptionStyle.European, name='put')
        portfolio = Portfolio(name='portfolio', priceables=[call, put])

        # Trade the position monthly without any scaling
        trade_action = AddScaledTradeAction(priceables=portfolio, trade_duration='1m',
                                            name='QuantityScaledAction1')
        trade_trigger = PeriodicTrigger(
            trigger_requirements=PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='1b'),
            actions=trade_action)

        strategy = Strategy(None, trade_trigger)

        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        summary = backtest.result_summary
        ledger = backtest.trade_ledger()

        assert len(summary) == 5
        assert len(ledger) == 10
        assert round(summary[Price].sum()) == 2715
        assert round(summary['Cumulative Cash'][-1]) == -922

        # Trade the position monthly and scale the quantity of the trade
        trade_action_scaled = AddScaledTradeAction(priceables=portfolio, trade_duration='1m',
                                                   scaling_level=scale_factor, name='QuantityScaledAction2')
        trade_trigger_scaled = PeriodicTrigger(
            trigger_requirements=PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='1b'),
            actions=trade_action_scaled)

        strategy_scaled = Strategy(None, trade_trigger_scaled)

        backtest_scaled = GE.run_backtest(strategy_scaled, start=start_date, end=end_date, frequency='1b',
                                          show_progress=True)

        summary_scaled = backtest_scaled.result_summary
        ledger_scaled = backtest_scaled.trade_ledger()

        # Price and cash should scale linearly
        assert len(summary_scaled) == len(summary)
        assert len(ledger_scaled) == len(ledger)
        assert round(summary_scaled[Price].sum()) == round(summary[Price].sum() * scale_factor)
        assert round(summary_scaled['Cumulative Cash'][-1]) == round(summary['Cumulative Cash'][-1] * scale_factor)


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_scaled_transaction_cost(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2024, 5, 6)
        end_date = dt.date(2024, 5, 10)

        # notional based transaction cost.  charge 1/10000th of the notional
        transaction_cost = ScaledTransactionModel('notional_amount', 0.0001)

        swap = IRSwap(notional_currency=Currency.GBP, notional_amount='50k', termination_date='1y', name='GBP1y')
        trade_action = AddTradeAction(priceables=swap, trade_duration='1m', name='Action1',
                                      transaction_cost=transaction_cost)
        trade_trigger = PeriodicTrigger(trigger_requirements=PeriodicTriggerRequirements(start_date=start_date,
                                                                                         end_date=end_date,
                                                                                         frequency='1b'),
                                        actions=trade_action)

        strategy = Strategy(None, trade_trigger)

        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        summary = backtest.result_summary
        ledger = backtest.trade_ledger()

        assert len(summary) == 5
        assert len(ledger) == 5
        assert round(summary[Price].sum()) == 90
        assert round(summary['Transaction Costs'][-1]) == 50000 * 0.0001 * 5 * -1


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_agg_transaction_cost(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2024, 5, 6)
        end_date = dt.date(2024, 5, 10)

        # aggregation based transaction cost.  charge 1/10000th of the notional + 1 for evey transaction
        models = tuple([ScaledTransactionModel('notional_amount', 0.0001), ConstantTransactionModel(1)])
        transaction_cost = AggregateTransactionModel(models, TransactionAggType.SUM)

        swap = IRSwap(notional_currency=Currency.GBP, notional_amount='50k', termination_date='1y', name='GBP1y')
        trade_action = AddTradeAction(priceables=swap, trade_duration='1m', name='Action1',
                                      transaction_cost=transaction_cost)
        trade_trigger = PeriodicTrigger(trigger_requirements=PeriodicTriggerRequirements(start_date=start_date,
                                                                                         end_date=end_date,
                                                                                         frequency='1b'),
                                        actions=trade_action)

        strategy = Strategy(None, trade_trigger)

        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        summary = backtest.result_summary
        ledger = backtest.trade_ledger()

        assert len(summary) == 5
        assert len(ledger) == 5
        assert round(summary[Price].sum()) == 90
        assert round(summary['Transaction Costs'][-1]) == (50000 * 0.0001 * 5 * -1) - 5


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_risk_scaled_transaction_cost(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2023, 6, 5)
        end_date = dt.date(2023, 6, 9)

        # risk based transaction cost.  Charge a 2 times the delta.
        transaction_cost = ScaledTransactionModel(IRDelta(aggregation_level='Type'), 2)

        swap = IRSwap(notional_currency=Currency.GBP, notional_amount='50k', termination_date='1y', name='GBP1y')
        trade_action = AddTradeAction(priceables=swap, trade_duration='1m', name='Action1',
                                      transaction_cost=transaction_cost)
        trade_trigger = PeriodicTrigger(trigger_requirements=PeriodicTriggerRequirements(start_date=start_date,
                                                                                         end_date=end_date,
                                                                                         frequency='1b'),
                                        actions=trade_action)

        strategy = Strategy(None, trade_trigger)

        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        summary = backtest.result_summary
        ledger = backtest.trade_ledger()

        assert len(summary) == 5
        assert len(ledger) == 5
        assert round(summary[Price].sum()) == -64
        assert round(summary['Transaction Costs'][-1]) == -62


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_hedge_transaction_costs(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2023, 6, 5)
        end_date = dt.date(2023, 6, 9)

        scaled_transaction_cost = ScaledTransactionModel('number_of_options', 1)
        fixed_transaction_cost = ConstantTransactionModel(0)

        opt = EqOption(underlier='.SPX', option_type=OptionType.Call, number_of_options=1, expiration_date='3m',
                       name='SPX_opt')
        trade_action = AddTradeAction(priceables=opt, trade_duration='1m', name='Action1',
                                      transaction_cost=fixed_transaction_cost)
        trade_trigger = PeriodicTrigger(trigger_requirements=PeriodicTriggerRequirements(start_date=start_date,
                                                                                         end_date=end_date,
                                                                                         frequency='1m'),
                                        actions=trade_action)
        hedge_port = Portfolio([EqOption(underlier='.SPX', option_type=OptionType.Call, number_of_options=1,
                                         expiration_date='3m', name='syn_fwd_call'),
                                EqOption(underlier='.SPX', option_type=OptionType.Put, number_of_options=1,
                                         expiration_date='3m', name='syn_fwd_put')])
        hedge_action = HedgeAction(risk=EqDelta, priceables=hedge_port, trade_duration='1b',
                                   transaction_cost=scaled_transaction_cost, name='HedgeAction1')
        hedge_trigger = PeriodicTrigger(trigger_requirements=PeriodicTriggerRequirements(start_date=start_date,
                                                                                         end_date=end_date,
                                                                                         frequency='1b'),
                                        actions=hedge_action)

        strategy = Strategy(None, [trade_trigger, hedge_trigger])
        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        summary = backtest.result_summary

        assert len(summary) == 5

        total_expected_transaction_costs = 0
        cur_bought_hedges = 0
        for d, port in backtest.portfolio_dict.items():
            # sale of prior hedges
            total_expected_transaction_costs += cur_bought_hedges
            cur_bought_hedges = sum(i.number_of_options for i in port.priceables if i.name.startswith('Scaled'))
            # buy next hedge
            total_expected_transaction_costs += cur_bought_hedges
            np.testing.assert_almost_equal(-summary['Transaction Costs'][d], total_expected_transaction_costs)


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_exit_transaction_costs(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2024, 6, 3)
        end_date = dt.date(2024, 6, 7)

        scaled_cost = 2
        scaled_transaction_cost = ScaledTransactionModel('number_of_options', scaled_cost)
        fixed_cost = 5
        fixed_transaction_cost = ConstantTransactionModel(fixed_cost)

        opt = EqOption(underlier='.SPX', option_type=OptionType.Call, number_of_options=1, expiration_date='3m',
                       name='SPX_opt')
        trade_action = AddTradeAction(priceables=opt, trade_duration='1b', transaction_cost=scaled_transaction_cost,
                                      transaction_cost_exit=fixed_transaction_cost, name='Action1')
        trade_trigger = PeriodicTrigger(trigger_requirements=PeriodicTriggerRequirements(start_date=start_date,
                                                                                         end_date=end_date,
                                                                                         frequency='1b'),
                                        actions=trade_action)

        strategy = Strategy(None, trade_trigger)
        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        summary = backtest.result_summary

        assert len(summary) == 5

        total_expected_transaction_costs = 0
        for d, port in backtest.portfolio_dict.items():
            # sale of prior trades
            total_expected_transaction_costs += 0 if d == start_date else fixed_cost
            cur_bought = sum(i.number_of_options for i in port.priceables)
            # buy next trade
            total_expected_transaction_costs += scaled_cost * cur_bought
            np.testing.assert_almost_equal(-summary['Transaction Costs'][d], total_expected_transaction_costs)


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_add_scaled_action_nav(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 6)
        end_date = dt.date(2021, 12, 10)

        initial_cash = 100

        # Define instruments for strategy
        call = EqOption('.STOXX50E', expiration_date='1m', strike_price='ATM', option_type=OptionType.Call,
                        option_style=OptionStyle.European, name='call')

        # NAV trading strategy with specified initial cash
        trade_action_scaled = AddScaledTradeAction(priceables=call, trade_duration='1b',
                                                   scaling_level=initial_cash,
                                                   scaling_type=ScalingActionType.NAV,
                                                   name='QuantityScaledAction1')

        trade_trigger_scaled = PeriodicTrigger(
            trigger_requirements=PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='1b'),
            actions=trade_action_scaled)

        strategy = Strategy(None, trade_trigger_scaled)

        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        summary = backtest.result_summary
        ledger = backtest.trade_ledger().to_dict('index')

        # Start with initial cash and only use sale proceeds to buy new options
        np.testing.assert_almost_equal(ledger['QuantityScaledAction1_call_2021-12-06']['Open Value'], -initial_cash)
        np.testing.assert_almost_equal(ledger['QuantityScaledAction1_call_2021-12-07']['Open Value'],
                                       -ledger['QuantityScaledAction1_call_2021-12-06']['Close Value'])
        np.testing.assert_almost_equal(ledger['QuantityScaledAction1_call_2021-12-08']['Open Value'],
                                       -ledger['QuantityScaledAction1_call_2021-12-07']['Close Value'])
        np.testing.assert_almost_equal(ledger['QuantityScaledAction1_call_2021-12-09']['Open Value'],
                                       -ledger['QuantityScaledAction1_call_2021-12-08']['Close Value'])
        np.testing.assert_almost_equal(ledger['QuantityScaledAction1_call_2021-12-10']['Open Value'],
                                       -ledger['QuantityScaledAction1_call_2021-12-09']['Close Value'])

        # Total cash spent is the initial cash throughout the entire strategy
        np.testing.assert_almost_equal(summary['Cumulative Cash'][0], -initial_cash)
        for c in summary['Cumulative Cash']:
            np.testing.assert_almost_equal(c, summary['Cumulative Cash'][0])


def nav_scaled_action_transaction_cost_test_for_agg_type(mocker, agg_type):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 6)
        end_date = dt.date(2021, 12, 10)

        initial_cash = 100

        # Define instruments for strategy
        call = EqOption('.STOXX50E', expiration_date='3m', strike_price='ATM', option_type=OptionType.Call,
                        option_style=OptionStyle.European, name='call')

        scaled_transaction_cost = ScaledTransactionModel(EqVega, 1.2)
        fixed_transaction_cost = ConstantTransactionModel(1)
        agg_tc = AggregateTransactionModel((scaled_transaction_cost, fixed_transaction_cost), agg_type)

        # NAV trading strategy with specified initial cash
        trade_action_scaled = AddScaledTradeAction(priceables=call, trade_duration='1b',
                                                   scaling_level=initial_cash,
                                                   scaling_type=ScalingActionType.NAV,
                                                   transaction_cost=agg_tc,
                                                   name=f'QuantityScaledAction2_{agg_type}')

        trade_trigger_scaled = PeriodicTrigger(
            trigger_requirements=PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='1b'),
            actions=trade_action_scaled)

        strategy = Strategy(None, trade_trigger_scaled)

        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        summary = backtest.result_summary

        ledger = backtest.trade_ledger().to_dict('index')
        tc = backtest.transaction_costs

        # Prior close value minus today's transaction costs equal today's value of options bought
        np.testing.assert_almost_equal(ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-06']['Open Value'],
                                       -initial_cash - tc[dt.date(2021, 12, 6)])
        np.testing.assert_almost_equal(ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-07']['Open Value'],
                                       -ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-06']['Close Value'] -
                                       tc[dt.date(2021, 12, 7)])
        np.testing.assert_almost_equal(ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-08']['Open Value'],
                                       -ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-07']['Close Value'] -
                                       tc[dt.date(2021, 12, 8)])
        np.testing.assert_almost_equal(ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-09']['Open Value'],
                                       -ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-08']['Close Value'] -
                                       tc[dt.date(2021, 12, 9)])
        np.testing.assert_almost_equal(ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-10']['Open Value'],
                                       -ledger[f'QuantityScaledAction2_{agg_type}_call_2021-12-09']['Close Value'] -
                                       tc[dt.date(2021, 12, 10)])

        # Cash spent on trades + Transaction costs are equal to the initial cash throughout the entire strategy
        total_cash_spent = summary['Cumulative Cash'] + summary['Transaction Costs']
        np.testing.assert_almost_equal(total_cash_spent[0], -initial_cash)
        for c in total_cash_spent:
            np.testing.assert_almost_equal(c, total_cash_spent[0])


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_add_scaled_action_nav_with_transaction_costs(mocker):
    for agg_type in (TransactionAggType.SUM, TransactionAggType.MAX, TransactionAggType.MIN):
        nav_scaled_action_transaction_cost_test_for_agg_type(mocker, agg_type)


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_generic_engine_custom_price_measure(mocker):
    with MockCalc(mocker):
        trig_date = dt.date(2021, 12, 1)
        call = EqOption('.STOXX50E', expiration_date='1y', strike_price='ATM', option_type=OptionType.Call,
                        option_style=OptionStyle.European, name='call')
        freq = '1m'
        trig_req = DateTriggerRequirements(dates=[trig_date])
        actions = AddTradeAction(call, freq, name='Action1')

        triggers = DateTrigger(trig_req, actions)
        strategy = Strategy(None, triggers)

        engine = GenericEngine(price_measure=DollarPrice)
        backtest = engine.run_backtest(strategy,
                                       states=[dt.date(2021, 12, 1), dt.date(2021, 12, 2), dt.date(2021, 12, 3)],
                                       show_progress=True)
        summary = backtest.result_summary
        assert len(summary) == 3
        assert round(summary[DollarPrice].sum()) == 804
        assert round(summary['Cumulative Cash'][-1]) == -291


def test_serialisation(mocker):
    call = FXOption(buy_sell='Buy', option_type='Call', pair='USDJPY', strike_price='ATMF',
                    notional_amount=1e5, expiration_date='2y', name='2y_call')
    hedge = FXForward(pair='USDJPY', settlement_date='2y', name='2y_hedge')
    d1 = dt.date(2023, 4, 11)
    d2 = dt.date(2023, 12, 25)
    d3 = dt.date(2024, 2, 29)
    dt1 = dt.datetime(2024, 4, 11, 9, 17, 34)
    dt2 = dt.datetime(2024, 4, 12, 10, 7, 57)
    sample_data_source = GsDataSource('DATASET_ABC', 'ASSET123', d1, d2)
    generic_data_source = GenericDataSource(s, MissingDataStrategy.fill_forward)
    # Actions to check
    add_trade_action_1 = AddTradeAction(call, '1m', name='Action1')
    add_trade_action_2 = AddTradeAction(call, d1, name='Action2')
    add_trade_action_3 = AddTradeAction(call, d2, name='Action3')
    hedge_action = HedgeAction(FXDelta(aggregation_level='type'), hedge, name='HedgeAction')
    exit_trade_action = ExitTradeAction('2y_call', name='ExitAction1')
    exit_all_trades_action = ExitAllPositionsAction(name='ExitEverything')
    add_scaled_trade = AddScaledTradeAction(priceables=call, trade_duration='1b',
                                            scaling_level=100,
                                            scaling_type=ScalingActionType.NAV,
                                            name='QuantityScaledAction1')

    # Triggers to check (randomly assign actions to triggers to cover all above actions)
    date_trigger = DateTrigger(DateTriggerRequirements(dates=(dt.date(2021, 12, 1),)), add_trade_action_1)
    periodic_trigger = PeriodicTrigger(PeriodicTriggerRequirements(d1, d2, "3m", (d3,)), add_trade_action_2)
    triggers = (
        date_trigger,
        periodic_trigger,
        DateTrigger(DateTriggerRequirements(dates=(dt.date(2021, 12, 1),)), [add_trade_action_3, hedge_action]),
        PeriodicTrigger(PeriodicTriggerRequirements(d1, d2, "3m", (d3,))),
        IntradayPeriodicTrigger(IntradayTriggerRequirements(dt1.time(), dt2.time(), 5.0)),
        MktTrigger(MktTriggerRequirements(generic_data_source, 1.1, TriggerDirection.BELOW), exit_trade_action),
        StrategyRiskTrigger(RiskTriggerRequirements(DollarPrice, -1.4, TriggerDirection.BELOW), exit_all_trades_action),
        AggregateTrigger(AggregateTriggerRequirements((date_trigger, periodic_trigger)), add_scaled_trade),
        NotTrigger(NotTriggerRequirements(date_trigger)),
        PortfolioTrigger(PortfolioTriggerRequirements('len', 2)),
        MeanReversionTrigger(MeanReversionTriggerRequirements(sample_data_source, 1.5, 0, 5)),
    )
    strategy = Strategy(None, triggers)

    # Test to_dict()
    strategy_dict = strategy.to_dict()
    reconstituted_strategy = Strategy.from_dict(strategy_dict)
    assert reconstituted_strategy == strategy

    # Test to_json()
    strategy_json = json.dumps(strategy.to_dict(), cls=JSONEncoder)
    reconstituted_strategy = Strategy.from_json(strategy_json)
    assert reconstituted_strategy == strategy


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_initial_portfolio(mocker):
    with MockCalc(mocker):
        irswap = IRSwap(PayReceive.Pay, '10y', Currency.USD, notional_amount=10000, name='10y')
        actions = AddTradeAction(irswap, '1b')

        new_ir_swap = IRSwap(PayReceive.Pay, '2y', Currency.USD, notional_amount=10000, name='2y')
        newer_ir_swap = IRSwap(PayReceive.Pay, '5y', Currency.USD, notional_amount=10000, name='5y')

        initial_port = {dt.date(2024, 5, 3): irswap,
                        dt.date(2024, 5, 24): [irswap, new_ir_swap],
                        dt.date(2024, 6, 14): [new_ir_swap, newer_ir_swap]}

        trig_req = PeriodicTriggerRequirements(start_date=dt.date(2024, 5, 3), end_date=dt.date(2024, 7, 5),
                                               frequency='1b')
        triggers = PeriodicTrigger(trig_req, actions)

        strategy = Strategy(initial_port, triggers)

        engine = GenericEngine()

        pricing_dates = [dt.date(2024, 5, 3),
                         dt.date(2024, 5, 10),
                         dt.date(2024, 5, 17),
                         dt.date(2024, 5, 24),
                         dt.date(2024, 5, 31),
                         dt.date(2024, 6, 7),
                         dt.date(2024, 6, 14),
                         dt.date(2024, 6, 21),
                         dt.date(2024, 6, 28),
                         dt.date(2024, 7, 5)]

        backtest = BackTest(strategy, pricing_dates, [Price], Price)
        engine._resolve_initial_portfolio(initial_port, backtest, dt.date(2024, 5, 3), pricing_dates, None)

        port_dict = backtest.portfolio_dict
        assert len(port_dict[dt.date(2024, 5, 3)]) == 1
        assert len(port_dict[dt.date(2024, 5, 24)]) == 2
        assert len(port_dict[dt.date(2024, 6, 14)]) == 2
        assert len(port_dict[dt.date(2024, 7, 5)]) == 2
        assert port_dict[dt.date(2024, 5, 24)][0].name == '10y_2024-05-24'
        assert port_dict[dt.date(2024, 5, 24)][1].name == '2y_2024-05-24'
        assert port_dict[dt.date(2024, 6, 14)][0].name == '2y_2024-06-14'
        assert port_dict[dt.date(2024, 6, 14)][1].name == '5y_2024-06-14'


@patch.object(GenericEngine, 'new_pricing_context', mock_pricing_context)
def test_add_scaled_trade_action_with_quantity_signal(mocker):
    with MockCalc(mocker):
        start_date = dt.date(2021, 12, 6)
        end_date = dt.date(2021, 12, 10)

        # Define instruments for strategy
        call = EqOption('.STOXX50E', expiration_date='3m', strike_price='ATM', option_type=OptionType.Call,
                        option_style=OptionStyle.European, name='call')

        trade_action_scaled = AddScaledTradeAction(priceables=call, trade_duration='1b',
                                                   scaling_level={start_date: 13, end_date: 21},
                                                   scaling_type=ScalingActionType.size,
                                                   name='AddScaledTradeAction1')

        trade_trigger_scaled = PeriodicTrigger(
            trigger_requirements=PeriodicTriggerRequirements(start_date=start_date, end_date=end_date, frequency='1b'),
            actions=trade_action_scaled)

        strategy = Strategy(None, trade_trigger_scaled)

        GE = GenericEngine()
        backtest = GE.run_backtest(strategy, start=start_date, end=end_date, frequency='1b', show_progress=True)

        portfolio_by_date = {d: p for d, p in backtest.portfolio_dict.items() if isinstance(d, dt.date)}
        assert all(len(p) == 1 for p in portfolio_by_date.values())
        quantity_by_date = {d: p.all_instruments[0].number_of_options for d, p in portfolio_by_date.items()}

        expected_quantity_by_date = {
            dt.date(2021, 12, 6): 13,
            dt.date(2021, 12, 7): 13,
            dt.date(2021, 12, 8): 13,
            dt.date(2021, 12, 9): 13,
            dt.date(2021, 12, 10): 21
        }

        assert quantity_by_date == expected_quantity_by_date



"""
Copyright 2021 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from unittest.mock import ANY

import pytest
from unittest import mock

from gs_quant.session import GsSession, Environment

from gs_quant.models.risk_model_utils import _upload_factor_data_if_present


@pytest.mark.parametrize('total_factors', [100])
def test__upload_factor_data_if_present(mocker, total_factors: int):

    from gs_quant.session import OAuth2Session
    OAuth2Session.init = mock.MagicMock(return_value=None)
    GsSession.use(Environment.QA, 'client_id', 'secret')

    date = "2024-03-28"
    factor_data = [{'factorCategory': '1', 'factorName': f'Factor {i + 1}', 'factorCategoryId': 'z',
                    'factorReturn': 0.001 * (i + 1)} for i in range(total_factors)]

    covariance_matrix = [[0 for i in range(total_factors)] for j in range(total_factors)]
    pre_vra_covariance_matrix = [[0 for i in range(total_factors)] for j in range(total_factors)]
    unadjusted_covariance_matrix = [[0 for i in range(total_factors)] for j in range(total_factors)]

    risk_model_data = {
        "factorData": factor_data,
        "covarianceMatrix": covariance_matrix,
        "preVRACovarianceMatrix": pre_vra_covariance_matrix,
        "unadjustedCovarianceMatrix": unadjusted_covariance_matrix,
    }

    # Ensuring GsSession._post receives the values that we pass to _upload_factor_data_if_present method
    def match_dictionaries(*args, **kwargs):
        url = args[0]
        actual_data = args[1]

        for key in risk_model_data:
            if key in ['preVRACovarianceMatrix', 'unadjustedCovarianceMatrix'] and 'awsUpload' not in url:
                # making sure new covariance matrices are only uploaded to AWS
                assert key not in actual_data
                continue
            assert key in actual_data
            assert risk_model_data[key] == actual_data[key]

    mocker.patch.object(GsSession.current, '_post', return_value='success', side_effect=match_dictionaries)

    _upload_factor_data_if_present('TEST_RISK_MODEL', risk_model_data, date)
    GsSession.current._post.assert_called_with('/risk/models/data/TEST_RISK_MODEL?partialUpload=true', ANY, timeout=200)

    _upload_factor_data_if_present('TEST_RISK_MODEL', risk_model_data, date, aws_upload=True)
    GsSession.current._post.assert_called_with('/risk/models/data/TEST_RISK_MODEL?partialUpload=true&awsUpload=true',
                                               ANY, timeout=200)
